# -*- coding: utf-8 -*-
from __future__ import division

from collections import defaultdict
from itertools import chain


def _imul(d1, d2):
    if hasattr(d2, 'keys'):
        for k in set(chain(d1.keys(), d2.keys())):
            d1[k] = d1[k]*d2[k]
    else:
        for k in d1:
            d1[k] *= d2


def _itruediv(d1, d2):
    if hasattr(d2, 'keys'):
        for k in set(chain(d1.keys(), d2.keys())):
            d1[k] = d1[k]/d2[k]
    else:
        for k in d1:
            d1[k] /= d2


class ArithmeticDict(defaultdict):
    """ A dictionary which supports arithmetics

    Subclassed from defaultdict, with support for addition, subtraction,
    multiplication and division. If other term/factor has a :meth:`keys` method
    the arithmetics are performed on a key per key basis. If :meth:`keys` is
    missing, the operation is broadcasted onto all values.
    Nonexisting keys are interpreted to signal a zero.

    Notes
    -----
    ``__eq__`` ignores values equal to ``self.default_factory()``

    Examples
    --------
    >>> d1 = ArithmeticDict(float, {'a': 2.0, 'b': 3.0})
    >>> d2 = ArithmeticDict(float, {'b': 5.0, 'c': 7.0})
    >>> (d1 + d2) == {'a': 2., 'b': 8., 'c': 7., 'd': 0.}
    True
    >>> (d1 * d1) == {'a': 4.0, 'b': 9.0, 'z': 0}
    True
    >>> (d1 * d2) == {'b': 15}
    True
    >>> d1*2 == {'a': 4, 'b': 6}
    True
    >>> (d1 / {'a': 2, 'b': 11})['b'] == 3./11
    True
    >>> d2/3 == {'b': 5./3, 'c': 7./3}
    True

    """

    def copy(self):
        return self.__class__(self.default_factory, self.items())

    def __iadd__(self, other):
        try:
            for k, v in other.items():
                self[k] += v
        except AttributeError:
            for k in self:
                self[k] += other
        return self

    def __isub__(self, other):
        try:
            for k, v in other.items():
                self[k] -= v
        except AttributeError:
            for k in self:
                self[k] -= other
        return self

    def __add__(self, other):
        a = self.copy()
        a += other
        return a

    def __sub__(self, other):
        a = self.copy()
        a -= other
        return a

    def __radd__(self, other):
        return self + other

    def __rsub__(self, other):
        return -1*self + other

    def __imul__(self, other):
        _imul(self, other)
        return self

    def __mul__(self, other):
        a = self.copy()
        a *= other
        return a

    def __rmul__(self, other):
        return self * other

    def __itruediv__(self, other):
        _itruediv(self, other)
        return self

    def __truediv__(self, other):
        a = self.copy()
        a /= other
        return a

    def __rtruediv__(self, other):
        """ other / self """
        return self.__class__(self.default_factory,
                              {k: other/v for k, v in self.items()})

    def __ifloordiv__(self, other):
        if hasattr(other, 'keys'):
            for k in set(chain(self.keys(), other.keys())):
                self[k] = self[k]//other[k]
        else:
            for k in self:
                self[k] //= other
        return self

    def __floordiv__(self, other):
        a = self.copy()
        a //= other
        return a

    def __rfloordiv__(self, other):
        """ other // self """
        return self.__class__(self.default_factory,
                              {k: other//v for k, v in self.items()})

    __div__ = __truediv__  # Py2 compatibility (or: import division from __future__)
    __idiv__ = __itruediv__  # Py2 compatibility
    __rdiv__ = __rtruediv__  # Py2 compatibility

    def __repr__(self):
        return "{}({}, {})".format(self.__class__.__name__,
                                   repr(self.default_factory),
                                   dict(self))

    def _element_eq(self, a, b):
        return a == b

    def _discrepancy(self, other, cb):
        default = self.default_factory()
        _self = self.copy()  # getitem is not idempotent on defaultdict
        _other = other.copy()
        try:
            for k in set(chain(_self.keys(), _other.keys())):
                if not cb(_self[k], _other.get(k, default)):
                    return False
            return True
        except TypeError:
            return False

    def __eq__(self, other):
        return self._discrepancy(other, self._element_eq)

    def isclose(self, other, rtol=1e-12, atol=None):
        def _isclose(a, b):
            lim = abs(rtol*b)
            if atol is not None:
                lim += atol
            return abs((a-b)) <= lim
        return self._discrepancy(other, _isclose)

    def all_non_negative(self):
        for v in self.values():
            if v < v*0:
                return False
        return True



# -*- coding: utf-8 -*-
import sys
from .string import StrPrinter
from .numbers import number_to_scientific_unicode


class UnicodePrinter(StrPrinter):

    _default_settings = dict(
        StrPrinter._default_settings,
        repr_name='unicode',
        Equilibrium_arrow=u'⇌',
        Reaction_arrow=u'→',
        magnitude_fmt=number_to_scientific_unicode,
        unit_fmt=lambda dim: (
            dim.unicode if sys.version_info[0] > 2
            else dim.unicode.decode(encoding='utf-8')
        )
    )
    _str = str if sys.version_info[0] > 2 else unicode  # noqa

    def _print_Substance(self, s, **kwargs):
        return s.unicode_name or s.name


def unicode_(obj, **settings):  # Python 2 keyword, hence the trailing '_'
    return UnicodePrinter(settings).doprint(obj)



# -*- coding: utf-8 -*-
"""
This package implements various parameterisations of properties from the
litterature with relevance in chemistry.
"""
from __future__ import (absolute_import, division, print_function)



# -*- coding: utf-8 -*-
"""
Non-public API (classes in this module may change without notice).

The purpose here is to define conventions, e.g. lower-case string
 'temperature' is used, opposed to e.g. 'T', 'Temperature', etc.
"""
from __future__ import (absolute_import, division, print_function)

from ..util._expr import create_Poly, create_Piecewise

TPoly = create_Poly('temperature')
RTPoly = create_Poly('temperature', reciprocal=True)
Log10TPoly = create_Poly('log10_temperature')
ShiftedTPoly = create_Poly('temperature', shift='Tref', name='ShiftedTPoly')
ShiftedLog10TPoly = create_Poly('log10_temperature', shift='log10_Tref')
ShiftedRTPoly = create_Poly('temperature', shift='Tref', reciprocal=True)
TPiecewise = create_Piecewise('temperature', nan_fallback=True)



# -*- coding: utf-8 -*-
"""
Contains functions for the `Arrhenius equation
<https://en.wikipedia.org/wiki/Arrhenius_equation>`_
(:func:`arrhenius_equation`) and a convenience fitting routine
(:func:`fit_arrhenius_equation`).
"""
from __future__ import (absolute_import, division, print_function)

from .._util import get_backend
from ..util.regression import least_squares
from ..util.pyutil import defaultnamedtuple
from ..units import default_constants, default_units, format_string

try:
    import numpy as np
except ImportError:
    np = None


def _fit_linearized(backtransfm, lin_x, lin_y, lin_yerr):
    if len(lin_x) != len(lin_y):
        raise ValueError("k and T needs to be of equal length.")
    if lin_yerr is not None:
        if len(lin_yerr) != len(lin_y):
            raise ValueError("kerr and T needs to be of equal length.")
    lin_p, lin_vcv, lin_r2 = least_squares(lin_x, lin_y, lin_yerr)
    return [cb(lin_p) for cb in backtransfm]


def _fit(T, k, kerr, func, lin_x, lin_y, backtransfm, linearized=False):
    _lin_y = lin_y(T, k)
    if kerr is None:
        lin_yerr = 1
    else:
        lin_yerr = (abs(lin_y(k - kerr, T) - _lin_y) +
                    abs(lin_y(k + kerr, T) - _lin_y))/2

    lopt = _fit_linearized(backtransfm, lin_x(T, k), _lin_y, lin_yerr)
    if linearized:
        return lopt
    from scipy.optimize import curve_fit
    popt, pcov = curve_fit(func, T, k, lopt, kerr)
    return popt, pcov


def _get_R(constants=None, units=None):
    if constants is None:
        R = 8.314472
        if units is not None:
            J = units.Joule
            K = units.Kelvin
            mol = units.mol
            R *= J/mol/K
    else:
        R = constants.molar_gas_constant
    return R


def arrhenius_equation(A, Ea, T, constants=None, units=None, backend=None):
    """
    Returns the rate coefficient according to the Arrhenius equation

    Parameters
    ----------
    A: float with unit
        frequency factor
    Ea: float with unit
        activation energy
    T: float with unit
        temperature
    constants: object (optional, default: None)
        if None:
            T assumed to be in Kelvin, Ea in J/(K mol)
        else:
            attributes accessed: molar_gas_constant
            Tip: pass quantities.constants
    units: object (optional, default: None)
        attributes accessed: Joule, Kelvin and mol
    backend: module (optional)
        module with "exp", default: numpy, math

    """
    be = get_backend(backend)
    R = _get_R(constants, units)
    try:
        RT = (R*T).rescale(Ea.dimensionality)
    except AttributeError:
        RT = R*T
    return A*be.exp(-Ea/RT)


def fit_arrhenius_equation(T, k, kerr=None, linearized=False, constants=None, units=None):
    """ Curve fitting of the Arrhenius equation to data points

    Parameters
    ----------
    T : float
    k : array_like
    kerr : array_like (optional)
    linearized : bool

    """
    return _fit(T, k, kerr, arrhenius_equation, lambda T, k: 1/T, lambda T, k: np.log(k),
                [lambda p: np.exp(p[0]), lambda p: -p[1]*_get_R(constants, units)], linearized=linearized)


def _fit_arrhenius_equation(T, k, kerr=None, linearized=False):
    """ Curve fitting of the Arrhenius equation to data points

    Parameters
    ----------
    k : array_like
    T : float
    kerr : array_like (optional)
    linearized : bool

    """
    if len(k) != len(T):
        raise ValueError("k and T needs to be of equal length.")
    from math import exp
    import numpy as np
    p = np.polyfit(1/T, np.log(k), 1)
    R = _get_R(constants=None, units=None)
    Ea = -R*p[0]
    A = exp(p[1])
    if linearized:
        return A, Ea
    from scipy.optimize import curve_fit
    if kerr is None:
        weights = None
    else:
        weights = 1/kerr**2
    popt, pcov = curve_fit(arrhenius_equation, T, k, [A, Ea], weights)
    return popt, pcov


class ArrheniusParam(defaultnamedtuple('ArrheniusParam', 'A Ea ref', [None])):
    """ Kinetic data in the form of an Arrhenius parameterisation

    Parameters
    ----------
    Ea: float
        activation energy
    A: float
        preexponential prefactor (Arrhenius type eq.)
    ref: object (default: None)
        arbitrary reference (e.g. string representing citation key)

    Examples
    --------
    >>> k = ArrheniusParam(1e13, 40e3)
    >>> '%.5g' % k(298.15)
    '9.8245e+05'

    """

    @classmethod
    def from_fit_of_data(cls, T, k, kerr=None, **kwargs):
        args, vcv = fit_arrhenius_equation(T, k, kerr)
        return cls(*args, **kwargs)

    def __call__(self, T, constants=None, units=None, backend=None):
        """ Evaluates the arrhenius equation for a specified state

        Parameters
        ----------
        T: float
        constants: module (optional)
        units: module (optional)
        backend: module (default: math)

        See also
        --------
        See :func:`chempy.arrhenius.arrhenius_equation`.

        """
        return arrhenius_equation(self.A, self.Ea, T, constants=constants,
                                  units=units, backend=backend)

    def Ea_over_R(self, constants, units, backend=None):
        return self.Ea/_get_R(constants, units)

    def as_RateExpr(self, unique_keys=None, constants=None, units=None, backend=None):
        from .rates import Arrhenius, MassAction
        args = [self.A, self.Ea_over_R(constants, units)]
        return MassAction(Arrhenius(args, unique_keys))

    def format(self, precision, tex=False):
        try:
            str_A, str_A_unit = format_string(self.A, precision, tex)
            str_Ea, str_Ea_unit = format_string(self.Ea, precision, tex)
        except Exception:
            str_A, str_A_unit = precision.format(self.A), '-'
            str_Ea, str_Ea_unit = precision.format(self.Ea), '-'
        return (str_A, str_A_unit), (str_Ea, str_Ea_unit)

    def equation_as_string(self, precision, tex=False):
        (str_A, str_A_unit), (str_Ea, str_Ea_unit) = self.format(precision, tex)
        if tex:
            return r"{}\exp \left(-\frac{{{}}}{{RT}} \right)".format(
                str_A, str_Ea + ' ' + str_Ea_unit), str_A_unit
        else:
            return "{}*exp(-{}/(R*T))".format(str_A, str_Ea + ' ' + str_Ea_unit), str_A_unit

    def __str__(self):
        return ' '.join(self.equation_as_string('%.5g'))


class ArrheniusParamWithUnits(ArrheniusParam):
    def __call__(self, state, constants=default_constants, units=default_units,
                 backend=None):
        """ See :func:`chempy.arrhenius.arrhenius_equation`. """
        return super(ArrheniusParamWithUnits, self).__call__(
            state, constants, units, backend)

    def as_RateExpr(self, unique_keys=None, constants=default_constants, units=default_units):
        return super(ArrheniusParamWithUnits, self).as_RateExpr(unique_keys, constants, units)



from .numbers import number_to_scientific_html
from .string import StrPrinter


def _html_clsname(key):
    return "chempy_" + key.replace(
        '+', 'plus').replace(
        '-', 'minus').replace(
        '(', 'leftparen').replace(
        ')', 'rightparen')

_html_semicolon = '&#59; '


class HTMLPrinter(StrPrinter):

    printmethod_attr = '_html'
    _default_settings = dict(
        StrPrinter._default_settings,
        repr_name='html',
        Equilibrium_arrow='&harr;',
        Reaction_arrow='&rarr;',
        Reaction_param_separator=_html_semicolon,
        magnitude_fmt=number_to_scientific_html
    )

    def _print_Substance(self, s, **kwargs):
        return s.html_name or s.name

    def _print_ReactionSystem(self, rsys, **kwargs):
        return super(HTMLPrinter, self)._print_ReactionSystem(rsys, **kwargs).replace('\n', '<br>\n')


def html(obj, **settings):
    return HTMLPrinter(settings).doprint(obj)


class CSSPrinter(HTMLPrinter):
    def _print_Substance(self, s, **kwargs):
        key = s.name
        name = s.html_name or key
        common_sty = 'border-radius: 5pt; padding: 0pt 3pt 0pt 3pt;'
        colors = self._get('colors', **kwargs)
        if key in colors:
            style = 'background-color:#%s; border: 1px solid #%s; %s' % (colors[key] + (common_sty,))
        else:
            style = common_sty
        fmt = '<span class="%s" style="%s">%s</span>'
        return fmt % (_html_clsname(key), style, name)

    def _tr_id(self, rsys, i):
        return 'chempy_%d_%d' % (id(rsys), i)

    def _print_ReactionSystem(self, rsys, **kwargs):
        sep = '</td><td style="text-align:left;">&nbsp;'
        around = '</td><td style="text-align:center;">', '</td><td style="text-align:left;">'
        # cf. https://github.com/jupyter/notebook/issues/2160#issuecomment-352216152
        row_template = '<tr class="%s"><td style="text-align:right;">%s</td></tr>'
        rows = [row_template % (self._tr_id(rsys, i), s) for i, s in enumerate(map(
            lambda r: self._print(r, Reaction_param_separator=sep, Reaction_around_arrow=around),
            rsys.rxns
        ))]
        tab_template = '<table class="chempy_ReactionSystem chempy_%d">%s%s</table>'
        header = '<tr><th style="text-align:center;" colspan="4">%s</th></tr>' % (rsys.name or '')
        return tab_template % (id(rsys), header, '\n\n'.join(rows))


def css(obj, **settings):
    return CSSPrinter(settings).doprint(obj)



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

from math import log10, floor

from ..units import html_of_unit, latex_of_unit, unicode_of_unit, to_unitless, unit_of
from ..util.parsing import _unicode_sup


def roman(num):
    """
    Examples
    --------
    >>> roman(4)
    'IV'
    >>> roman(17)
    'XVII'

    """
    tokens = 'M CM D CD C XC L XL X IX V IV I'.split()
    values = 1000, 900, 500, 400, 100, 90, 50, 40, 10, 9, 5, 4, 1
    result = ''
    for t, v in zip(tokens, values):
        cnt = num//v
        result += t*cnt
        num -= v*cnt
    return result


def _mag(num):
    return int(floor(log10(abs(num))))


def _float_str_w_uncert(x, xe, precision=2):
    """ Prints uncertain number with parenthesis

    Parameters
    ----------
    x : nominal value
    xe : uncertainty
    precision : number of significant digits in uncertainty

    Examples
    --------
    >>> _float_str_w_uncert(-9.99752e5, 349, 3)
    '-999752(349)'
    >>> _float_str_w_uncert(-9.99752e15, 349e10, 2)
    '-9.9975(35)e15'
    >>> _float_str_w_uncert(3.1416, 0.029, 1)
    '3.14(3)'
    >>> _float_str_w_uncert(3.1416e9, 2.9e6, 1)
    '3.142(3)e9'

    Returns
    -------
    shortest string representation of "x +- xe" either as
    ``x.xx(ee)e+xx`` or ``xxx.xx(ee)``

    Notes
    -----
    The code in this function is from a question on StackOverflow:
        http://stackoverflow.com/questions/6671053
        written by:
            Lemming, http://stackoverflow.com/users/841562/lemming
        the code is licensed under 'CC-WIKI'.
        (see: http://blog.stackoverflow.com/2009/06/attribution-required/)

    """
    # base 10 exponents
    x_exp = int(floor(log10(abs(x))))
    xe_exp = int(floor(log10(abs(xe))))

    # uncertainty
    un_exp = xe_exp-precision+1
    un_int = round(xe*10**(-un_exp))

    # nominal value
    no_exp = un_exp
    no_int = round(x*10**(-no_exp))

    # format - nom(unc)exp
    fieldw = x_exp - no_exp
    fmt = '%%.%df' % fieldw
    result1 = (fmt + '(%.0f)e%d') % (no_int*10**(-fieldw), un_int, x_exp)

    # format - nom(unc)
    fieldw = max(0, -no_exp)
    fmt = '%%.%df' % fieldw
    result2 = (fmt + '(%.0f)') % (no_int*10**no_exp, un_int*10**max(0, un_exp))

    # return shortest representation
    if len(result2) <= len(result1):
        return result2
    else:
        return result1


def _number_to_X(number, uncertainty, unit, fmt, unit_fmt, fmt_pow_10, space=' '):
    uncertainty = uncertainty or getattr(number, 'uncertainty', None)
    unit = unit or unit_of(number)
    if unit is 1:
        unit_str = ''
        mag = number
    else:
        unit_str = space + unit_fmt(unit)
        mag = to_unitless(number, unit)
        if uncertainty is not None:
            uncertainty = to_unitless(uncertainty, unit)

    if uncertainty is None:
        if fmt is None:
            fmt = 5
        if isinstance(fmt, int):
            flt = ('%%.%dg' % fmt) % mag
        else:
            flt = fmt(mag)
    else:
        if fmt is None:
            fmt = 2
        if isinstance(fmt, int):
            flt = _float_str_w_uncert(mag, uncertainty, fmt)
        else:
            flt = fmt(mag, uncertainty)
    if 'e' in flt:
        significand, mantissa = flt.split('e')
        return fmt_pow_10(significand, mantissa) + unit_str
    else:
        return flt + unit_str


def _latex_pow_10(significand, mantissa):
    if significand in ('1', '1.0'):
        fmt = '10^{%s}'
    else:
        fmt = significand + r'\cdot 10^{%s}'
    return fmt % str(int(mantissa))


def number_to_scientific_latex(number, uncertainty=None, unit=None, fmt=None):
    r""" Formats a number as LaTeX (optionally with unit/uncertainty)

    Parameters
    ----------
    number : float (w or w/o unit)
    uncertainty : same as number
    unit : unit
    fmt : int or callable

    Examples
    --------
    >>> number_to_scientific_latex(3.14) == '3.14'
    True
    >>> number_to_scientific_latex(3.14159265e-7)
    '3.1416\\cdot 10^{-7}'
    >>> import quantities as pq
    >>> number_to_scientific_latex(2**0.5 * pq.m / pq.s)
    '1.4142\\,\\mathrm{\\frac{m}{s}}'
    >>> number_to_scientific_latex(1.23456, .789, fmt=2)
    '1.23(79)'

    """
    return _number_to_X(number, uncertainty, unit, fmt, latex_of_unit, _latex_pow_10, '\,')


def _unicode_pow_10(significand, mantissa):
    if significand in ('1', '1.0'):
        result = u'10'
    else:
        result = significand + u'·10'
    return result + u''.join(map(_unicode_sup.get, str(int(mantissa))))


def number_to_scientific_unicode(number, uncertainty=None, unit=None, fmt=None):
    u""" Formats a number as unicode (optionally with unit/uncertainty)

    Parameters
    ----------
    number : float (w or w/o unit)
    uncertainty : same as number
    unit : unit
    fmt : int or callable

    Examples
    --------
    >>> number_to_scientific_unicode(3.14) == u'3.14'
    True
    >>> number_to_scientific_unicode(3.14159265e-7) == u'3.1416·10⁻⁷'
    True
    >>> import quantities as pq
    >>> number_to_scientific_unicode(2**0.5 * pq.m / pq.s)
    '1.4142 m/s'

    """
    return _number_to_X(number, uncertainty, unit, fmt, unicode_of_unit, _unicode_pow_10)


def _html_pow_10(significand, mantissa):
    if significand in ('1', '1.0'):
        result = '10<sup>'
    else:
        result = significand + '&sdot;10<sup>'
    return result + str(int(mantissa)) + '</sup>'


def number_to_scientific_html(number, uncertainty=None, unit=None, fmt=None):
    r""" Formats a number as HTML (optionally with unit/uncertainty)

    Parameters
    ----------
    number : float (w or w/o unit)
    uncertainty : same as number
    unit : unit
    fmt : int or callable

    Examples
    --------
    >>> number_to_scientific_html(3.14) == '3.14'
    True
    >>> number_to_scientific_html(3.14159265e-7)
    '3.1416&sdot;10<sup>-7</sup>'
    >>> number_to_scientific_html(1e13)
    '10<sup>13</sup>'
    >>> import quantities as pq
    >>> number_to_scientific_html(2**0.5 * pq.m / pq.s)
    '1.4142 m/s'

    """
    return _number_to_X(number, uncertainty, unit, fmt, html_of_unit, _html_pow_10)



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

from .numbers import number_to_scientific_html, number_to_scientific_latex, number_to_scientific_unicode
from .table import as_per_substance_html_table
from .js import javascript
from .string import str_
from .pretty import unicode_
from .web import html, css
from .tex import latex



# -*- coding: utf-8 -*-

symbols = (
    'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al',
    'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe',
    'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr',
    'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn',
    'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm',
    'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W',
    'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn',
    'Fr', 'Ra', 'Ac', 'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf',
    'Es', 'Fm', 'Md', 'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds',
    'Rg', 'Cn', 'Uut', 'Fl', 'Uup', 'Lv', 'Uus', 'Uuo'
)

period_lengths = (2, 8, 8, 18, 18, 32, 32)
accum_period_lengths = (2, 10, 18, 36, 54, 86, 118)

# icosagens, crystallogens, pnictogens, chalcogens, halogens
groups = {g: tuple(x - 18 + g for x in accum_period_lengths[1:]) for g in range(13, 18)}
groups[1] = (1,) + tuple(x + 1 for x in accum_period_lengths[:-1])  # alkali metals
groups[2] = tuple(x + 2 for x in accum_period_lengths[:-1])  # alkaline earth metals
groups[18] = accum_period_lengths  # noble gases

names = (
    'Hydrogen', 'Helium', 'Lithium', 'Beryllium',
    'Boron', 'Carbon', 'Nitrogen', 'Oxygen', 'Fluorine', 'Neon', 'Sodium',
    'Magnesium', 'Aluminium', 'Silicon', 'Phosphorus', 'Sulfur',
    'Chlorine', 'Argon', 'Potassium', 'Calcium', 'Scandium', 'Titanium',
    'Vanadium', 'Chromium', 'Manganese', 'Iron', 'Cobalt', 'Nickel',
    'Copper', 'Zinc', 'Gallium', 'Germanium', 'Arsenic', 'Selenium',
    'Bromine', 'Krypton', 'Rubidium', 'Strontium', 'Yttrium', 'Zirconium',
    'Niobium', 'Molybdenum', 'Technetium', 'Ruthenium', 'Rhodium',
    'Palladium', 'Silver', 'Cadmium', 'Indium', 'Tin', 'Antimony',
    'Tellurium', 'Iodine', 'Xenon', 'Caesium', 'Barium', 'Lanthanum',
    'Cerium', 'Praseodymium', 'Neodymium', 'Promethium', 'Samarium',
    'Europium', 'Gadolinium', 'Terbium', 'Dysprosium', 'Holmium',
    'Erbium', 'Thulium', 'Ytterbium', 'Lutetium', 'Hafnium', 'Tantalum',
    'Tungsten', 'Rhenium', 'Osmium', 'Iridium', 'Platinum', 'Gold',
    'Mercury', 'Thallium', 'Lead', 'Bismuth', 'Polonium', 'Astatine',
    'Radon', 'Francium', 'Radium', 'Actinium', 'Thorium', 'Protactinium',
    'Uranium', 'Neptunium', 'Plutonium', 'Americium', 'Curium',
    'Berkelium', 'Californium', 'Einsteinium', 'Fermium', 'Mendelevium',
    'Nobelium', 'Lawrencium', 'Rutherfordium', 'Dubnium', 'Seaborgium',
    'Bohrium', 'Hassium', 'Meitnerium', 'Darmstadtium', 'Roentgenium',
    'Copernicium', '(Ununtrium)', 'Flerovium', '(Ununpentium)',
    'Livermorium', '(Ununseptium)', '(Ununoctium)'
)

lower_names = tuple(n.lower().lstrip('(').rstrip(')') for n in names)


def atomic_number(name):
    try:
        return symbols.index(name) + 1
    except ValueError:
        return lower_names.index(name.lower()) + 1

# The data in '_relative_atomic_masses' is licensed under the CC-SA license
# https://en.wikipedia.org/w/index.php?title=List_of_elements&oldid=700476748
_relative_atomic_masses = (
    "1.008 4.002602(2) 6.94 9.0121831(5) 10.81 12.011 14.007 15.999"
    " 18.998403163(6) 20.1797(6) 22.98976928(2) 24.305 26.9815385(7) 28.085"
    " 30.973761998(5) 32.06 35.45 39.948(1) 39.0983(1) 40.078(4)"
    " 44.955908(5) 47.867(1) 50.9415(1) 51.9961(6) 54.938044(3) 55.845(2)"
    " 58.933194(4) 58.6934(4) 63.546(3) 65.38(2) 69.723(1) 72.630(8)"
    " 74.921595(6) 78.971(8) 79.904 83.798(2) 85.4678(3) 87.62(1)"
    " 88.90584(2) 91.224(2) 92.90637(2) 95.95(1) [98] 101.07(2) 102.90550(2)"
    " 106.42(1) 107.8682(2) 112.414(4) 114.818(1) 118.710(7) 121.760(1)"
    " 127.60(3) 126.90447(3) 131.293(6) 132.90545196(6) 137.327(7)"
    " 138.90547(7) 140.116(1) 140.90766(2) 144.242(3) [145] 150.36(2)"
    " 151.964(1) 157.25(3) 158.92535(2) 162.500(1) 164.93033(2) 167.259(3)"
    " 168.93422(2) 173.045(10) 174.9668(1) 178.49(2) 180.94788(2) 183.84(1)"
    " 186.207(1) 190.23(3) 192.217(3) 195.084(9) 196.966569(5) 200.592(3)"
    " 204.38 207.2(1) 208.98040(1) [209] [210] [222] [223] [226] [227]"
    " 232.0377(4) 231.03588(2) 238.02891(3) [237] [244] [243] [247] [247]"
    " [251] [252] [257] [258] [259] [266] [267] [268] [269] [270] [269]"
    " [278] [281] [282] [285] [286] [289] [289] [293] [294] [294]"
)


def _get_relative_atomic_masses():
    for mass in _relative_atomic_masses.split():
        if mass.startswith('[') and mass.endswith(']'):
            yield float(mass[1:-1])
        elif '(' in mass:
            yield float(mass.split('(')[0])
        else:
            yield(float(mass))

relative_atomic_masses = tuple(_get_relative_atomic_masses())


def mass_from_composition(composition):
    """ Calculates molecular mass from atomic weights

    Parameters
    ----------
    composition: dict
        Dictionary mapping int (atomic number) to int (coefficient)

    Returns
    -------
    float
        molecular weight in atomic mass units


    Notes
    -----
    Atomic number 0 denotes charge or "net electron defficiency"

    Examples
    --------
    >>> '%.2f' % mass_from_composition({0: -1, 1: 1, 8: 1})
    '17.01'
    """
    mass = 0.0
    for k, v in composition.items():
        if k == 0:  # electron
            mass -= v*5.489e-4
        else:
            mass += v*relative_atomic_masses[k-1]
    return mass



"""
This module collects a few analytic solutions of initial value problems (IVPs) in
chemical kinetics (i.e. integrated rate expressions in closed form).
The expressions are useful in e.g. regression or for comparison with numerical
solution of the corresponding ODE system.
"""
from __future__ import (absolute_import, division, print_function)

from .._util import get_backend


def dimerization_irrev(t, kf, initial_C, P0=1, t0=0):
    return 1/(1/initial_C + 2*kf*(t-t0))


def pseudo_irrev(t, kf, prod, major, minor, backend=None):
    """ Analytic product transient of a irreversible pseudo first order reaction.

    Product concentration vs time from pseudo-first order irreversible kinetics.

    Parameters
    ----------
    t : float, Symbol or array_like
        Time.
    kf : number or Symbol
        Forward (bimolecular) rate constant.
    kb : number or Symbol
        Backward (unimolecular) rate constant.
    prod : number or Symbol
        Initial concentration of the complex.
    major : number or Symbol
        Initial concentration of the more abundant reactant.
    minor : number or Symbol
        Initial concentration of the less abundant reactant.
    backend : module or str
        Default is 'numpy', can also be e.g. ``sympy``.

    """
    be = get_backend(backend)
    return prod + minor*(1 - be.exp(-major*kf*t))
pseudo_irrev.name = 'Pseudo first order irreversible'


def pseudo_rev(t, kf, kb, prod, major, minor, backend=None):
    """ Analytic product transient of a reversible pseudo first order reaction.

    Product concentration vs time from pseudo-first order reversible kinetics.

    Parameters
    ----------
    t : float, Symbol or array_like
        Time.
    kf : number or Symbol
        Forward (bimolecular) rate constant.
    kb : number or Symbol
        Backward (unimolecular) rate constant.
    prod : number or Symbol
        Initial concentration of the complex.
    major : number or Symbol
        Initial concentration of the more abundant reactant.
    minor : number or Symbol
        Initial concentration of the less abundant reactant.
    backend : module or str
        Default is 'numpy', can also be e.g. ``sympy``.

    """
    be = get_backend(backend)
    return (
        -kb*prod + kf*major*minor + (kb*prod - kf*major*minor) *
        be.exp(-t*(kb + kf*major))
    )/(kb + kf*major)
pseudo_rev.name = 'Pseudo first order reversible'


def binary_irrev(t, kf, prod, major, minor, backend=None):
    """ Analytic product transient of a irreversible 2-to-1 reaction.

    Product concentration vs time from second order irreversible kinetics.

    Parameters
    ----------
    t : float, Symbol or array_like
    kf : number or Symbol
        Forward (bimolecular) rate constant.
    prod : number or Symbol
        Initial concentration of the complex.
    major : number or Symbol
        Initial concentration of the more abundant reactant.
    minor : number or Symbol
        Initial concentration of the less abundant reactant.
    backend : module or str
        Default is 'numpy', can also be e.g. ``sympy``.

    """
    be = get_backend(backend)
    return prod + major*(1 - be.exp(-kf*(major-minor)*t))/(
        major/minor - be.exp(-kf*t*(major-minor)))
binary_irrev.name = 'Second order irreversible'


def binary_rev(t, kf, kb, prod, major, minor, backend=None):
    """ Analytic product transient of a reversible 2-to-1 reaction.

    Product concentration vs time from second order reversible kinetics.

    Parameters
    ----------
    t : float, Symbol or array_like
        Time.
    kf : number or Symbol
        Forward (bimolecular) rate constant.
    kb : number or Symbol
        Backward (unimolecular) rate constant.
    prod : number or Symbol
        Initial concentration of the complex.
    major : number or Symbol
        Initial concentration of the more abundant reactant.
    minor : number or Symbol
        Initial concentration of the less abundant reactant.
    backend : module or str
        Default is 'numpy', can also be e.g. ``sympy``.

    """
    # see _integrated.ipynb for derivation
    be = get_backend(backend)
    X, Y, Z = prod, major, minor
    x0 = Y*kf
    x1 = Z*kf
    x2 = 2*X*kf
    x3 = -kb - x0 - x1
    x4 = -x2 + x3
    x5 = be.sqrt(-4*kf*(X**2*kf + X*x0 + X*x1 + Z*x0) + x4**2)
    x6 = kb + x0 + x1 + x5
    x7 = (x3 + x5)*be.exp(-t*x5)
    x8 = x3 - x5
    return (x4*x8 + x5*x8 + x7*(x2 + x6))/(2*kf*(x6 + x7))

binary_rev.name = 'Second order reversible'


def unary_irrev_cstr(t, k, r, p, fr, fp, fv, backend=None):
    """ Analytic solution for ``A -> B`` in a CSTR.

    Analytic solution for a first order process in a continuously
    stirred tank reactor (CSTR).

    Parameters
    ----------
    t : array_like
    k : float_like
        Rate constant
    r : float_like
        Initial concentration of reactant.
    p : float_like
        Initial concentration of product.
    fr : float_like
        Concentration of reactant in feed.
    fp : float_like
        Concentration of product in feed.
    fv : float_like
        Feed rate / tank volume ratio.
    backend : module or str
        Default is 'numpy', can also be e.g. ``sympy``.

    Returns
    -------
    length-2 tuple
        concentrations of reactant and product

    """
    # See _kinetics_cstr.ipynb
    be = get_backend(backend)
    x0 = fr*fv
    x1 = fv + k
    x2 = 1/x1
    x3 = fv*r + k*r - x0
    x4 = fr*k
    x5 = be.exp(-fv*t)
    return (
        x0*x2 + x2*x3*be.exp(-t*x1),
        -x2*x3*x5*(-1 + be.exp(-k*t)) + x2*x5*(-fp*fv - fp*k + fv*p + k*p - x4) + x2*(fp*x1 + x4)
    )


def binary_irrev_cstr(t, k, r, p, fr, fp, fv, n=1, backend=None):
    """ Analytic solution for ``2 A -> n B`` in a CSTR.

    Parameters
    ----------
    t : array_like
    k : float_like
        Rate constant
    r : float_like
        Initial concentration of reactant.
    p : float_like
        Initial concentration of product.
    fr : float_like
        Concentration of reactant in feed.
    fp : float_like
        Concentration of product in feed.
    fv : float_like
        Feed rate / tank volume ratio.
    n : int
    backend : module or str
        Default is 'numpy', can also be e.g. ``sympy``.

    Returns
    -------
    length-2 tuple
        concentrations of reactant and product

    """
    # Mathematica source:
    # FortranForm[
    # DSolve[{Derivative[1][A][t] == a0 f - f A[t] - 2 k A[t]^2,
    #   Derivative[1][B][t] == b0 f + n*k A[t]^2 - f B[t], A[0] == x,
    #   B[0] == y}, {A[t], B[t]}, {t}]]
    # Post processed using sympy's cse function.
    # (see _derive_analytic_cstr_bireac.ipynb)
    be = get_backend(backend)
    atanh = getattr(be, 'atanh', be.arctanh)
    three = 3*be.cos(0)

    x0 = 1/k
    x1 = be.sqrt(fv)
    x2 = 8*k
    x3 = fr*x2
    x4 = be.sqrt(fv + x3)
    x5 = x1*x4
    x6 = x1*x4/2
    x7 = atanh((-fv**(three/2)*x4 - 4*k*r*x5)/(fv**2 + fv*x3))
    x8 = fv*t
    x9 = fp*x2
    x10 = 4*k*n
    x11 = fr*x10
    x12 = be.exp(x8)
    x13 = n*x12
    return (
        x0*(-fv + x5*be.tanh(t*x6 - x7))/4,
        x0*(fv*x13 + 8*k*p + r*x10 - x1*x13*x4*be.tanh(
            x6*(t - 2*x7/(x1*x4))
        ) + x11*x12 - x11 + x12*x9 - x9)*be.exp(-x8)/8
    )



# -*- coding: utf-8 -*-
"""
This modules collects expressions related to ionic strenght, e.g. the Debye-Hückel expressions.
"""
from __future__ import (absolute_import, division, print_function)

from collections import OrderedDict
import warnings

from ._util import get_backend
from .chemistry import Substance
from .units import allclose


def _get_b0(b0, units=None):
    if units is not None and b0 is 1:
        return b0*units.molal
    else:
        return b0


def ionic_strength(molalities, charges=None, units=None, substances=None,
                   substance_factory=Substance.from_formula, warn=True):
    """ Calculates the ionic strength

    Parameters
    ----------
    molalities: array_like or dict
        Optionally with unit (amount / mass).
        when dict: mapping substance key to molality.
    charges: array_like
        Charge of respective ion, taken for substances when None.
    units: object (optional, default: None)
        Attributes accessed: molal.
    substances: dict, optional
        Mapping of substance keys to Substance instances (used when molalities
        is a dict).
    substance_factory: callback
        Used if `substances` is a string.
    warn: bool
        Issue a warning if molalities violates net charge neutrality.

    Examples
    --------
    >>> ionic_strength([1e-3, 3e-3], [3, -1]) == .5 * (9 + 3) * 1e-3
    True
    >>> ionic_strength({'Mg+2': 6, 'PO4-3': 4})
    30.0

    """
    tot = None
    if charges is None:
        if substances is None:
            substances = ' '.join(molalities.keys())
        if isinstance(substances, str):
            substances = OrderedDict([(k, substance_factory(k)) for k
                                      in substances.split()])
        charges, molalities = zip(*[(substances[k].charge, v) for k, v in molalities.items()])
    if len(molalities) != len(charges):
        raise ValueError("molalities and charges of different lengths")
    for b, z in zip(molalities, charges):
        if tot is None:
            tot = b * z**2
        else:
            tot += b * z**2
    if warn:
        net = None
        for b, z in zip(molalities, charges):
            if net is None:
                net = b * z
            else:
                net += b * z
        if not allclose(net, tot*0, atol=tot*1e-14):
            warnings.warn("Molalities not charge neutral: %s" % str(net))
    return tot/2


class _ActivityProductBase(object):
    """ Baseclass for activity products """

    def __init__(self, stoich, *args):
        self.stoich = stoich
        self.args = args

    def __call__(self, c):
        pass


def A(eps_r, T, rho, b0=1, constants=None, units=None, backend=None):
    """
    Debye Huckel constant A

    Parameters
    ----------
    eps_r: float
        relative permittivity
    T: float with unit
        Temperature (default: assume Kelvin)
    rho: float with unit
        density (default: assume kg/m**3)
    b0: float, optional
        Reference molality, optionally with unit (amount / mass)
        IUPAC defines it as 1 mol/kg. (default: 1).
    units: object (optional, default: None)
        attributes accessed: meter, Kelvin and mol
    constants: object (optional, default: None)
        if None:
            T assumed to be in Kelvin and b0 = 1 mol/kg
        else:
            see source code for what attributes are used.
            Tip: pass quantities.constants

    Notes
    -----
    Remember to divide by ln(10) if you want to use the constant
    with log10 based expression.

    References
    ----------
    Atkins, De Paula, Physical Chemistry, 8th edition

    """
    b0 = _get_b0(b0, units)
    be = get_backend(backend)
    one = be.pi**0
    if constants is None:
        combined = 132871.85866393594
        if units is not None:
            m = units.meter
            K = units.Kelvin
            mol = units.mol
            combined *= (m*K)**(3*one/2) / mol**(one/2)
        return combined*(rho * b0 * T**-3 * eps_r**-3)**0.5
    F = constants.Faraday_constant
    NA = constants.Avogadro_constant
    eps0 = constants.vacuum_permittivity
    kB = constants.Boltzmann_constant
    pi = constants.pi
    A = F**3/(4*pi*NA)*(rho*b0/(2*(eps0*eps_r*kB*NA*T)**3))**(one/2)
    return A


def B(eps_r, T, rho, b0=1, constants=None, units=None, backend=None):
    """
    Extended Debye-Huckel parameter B

    Parameters
    ----------
    eps_r: float
        relative permittivity
    T: float with unit
        temperature
    rho: float with unit
        density
    b0: float with unit
        reference molality
    units: object (optional, default: None)
        attributes accessed: meter, Kelvin and mol
    constants: object (optional, default: None)
        if None:
            T assumed to be in Kelvin, rho in kg/m**3 and b0 = 1 mol/kg
        else:
            attributes accessed: molar_gas_constant, Faraday_constant
            Tip: pass quantities.constants

    Returns
    -------
    Debye Huckel B constant (default in m**-1)

    """
    b0 = _get_b0(b0, units)
    be = get_backend(backend)
    one = be.pi**0
    if constants is None:
        combined = 15903203868.740343
        if units is not None:
            m = units.meter
            K = units.Kelvin
            mol = units.mol
            combined *= (m*K/mol)**(one/2)
        return combined*(rho*b0/(T*eps_r))**0.5
    F = constants.Faraday_constant
    eps0 = constants.vacuum_permittivity
    R = constants.molar_gas_constant
    B = F*(2*rho*b0/(eps_r*eps0*R*T))**(one/2)
    return B


def limiting_log_gamma(I, z, A, I0=1, backend=None):
    """ Debye-Hyckel limiting formula """
    be = get_backend(backend)
    one = be.pi**0
    return -A*z**2*(I/I0)**(one/2)


def extended_log_gamma(I, z, a, A, B, C=0, I0=1, backend=None):
    """ Debye-Huckel extended formula """
    be = get_backend(backend)
    one = be.pi**0
    I_I0 = I/I0
    sqrt_I_I0 = (I_I0)**(one/2)
    return -A*z**2 * sqrt_I_I0/(1 + B*a*sqrt_I_I0) + C*I_I0


def davies_log_gamma(I, z, A, C=-0.3, I0=1, backend=None):
    """ Davies formula """
    be = get_backend(backend)
    one = be.pi**0
    I_I0 = I/I0
    sqrt_I_I0 = (I_I0)**(one/2)
    return -A * z**2 * (sqrt_I_I0/(1 + sqrt_I_I0) + C*I_I0)


def limiting_activity_product(I, stoich, z, T, eps_r, rho, backend=None):
    """ Product of activity coefficients based on DH limiting law. """
    be = get_backend(backend)
    Aval = A(eps_r, T, rho)
    tot = 0
    for idx, nr in enumerate(stoich):
        tot += nr*limiting_log_gamma(I, z[idx], Aval)
    return be.exp(tot)


def extended_activity_product(I, stoich, z, a, T, eps_r, rho, C=0, backend=None):
    be = get_backend(backend)
    Aval = A(eps_r, T, rho)
    Bval = B(eps_r, T, rho)
    tot = 0
    for idx, nr in enumerate(stoich):
        tot += nr*extended_log_gamma(I, z[idx], a[idx], Aval, Bval, C)
    return be.exp(tot)


def davies_activity_product(I, stoich, z, a, T, eps_r, rho, C=-0.3, backend=None):
    be = get_backend(backend)
    Aval = A(eps_r, T, rho)
    tot = 0
    for idx, nr in enumerate(stoich):
        tot += nr*davies_log_gamma(I, z[idx], Aval, C)
    return be.exp(tot)


class LimitingDebyeHuckelActivityProduct(_ActivityProductBase):

    def __call__(self, c):
        z = self.args[0]
        I = ionic_strength(c, z)
        return limiting_activity_product(I, self.stoich, *self.args)


class ExtendedDebyeHuckelActivityProduct(_ActivityProductBase):

    def __call__(self, c):
        z = self.args[0]
        I = ionic_strength(c, z)
        return extended_activity_product(I, self.stoich, *self.args)



import warnings
from ..water_viscosity_korson_1969 import water_viscosity


def test_water_viscosity():
    warnings.filterwarnings("error")  # Table II (p. 38):
    assert abs(water_viscosity(273.15 +   0) - 1.7916) < 5e-4
    assert abs(water_viscosity(273.15 +   5) - 1.5192) < 5e-4
    assert abs(water_viscosity(273.15 +  10) - 1.3069) < 5e-4
    assert abs(water_viscosity(273.15 +  15) - 1.1382) < 5e-4
    assert abs(water_viscosity(273.15 +  20) - 1.0020) < 5e-4
    assert abs(water_viscosity(273.15 +  25) - 0.8903) < 5e-4
    assert abs(water_viscosity(273.15 +  30) - 0.7975) < 5e-4
    assert abs(water_viscosity(273.15 +  35) - 0.7195) < 5e-4
    assert abs(water_viscosity(273.15 +  40) - 0.6532) < 5e-4
    assert abs(water_viscosity(273.15 +  45) - 0.5963) < 5e-4
    assert abs(water_viscosity(273.15 +  50) - 0.5471) < 5e-4
    assert abs(water_viscosity(273.15 +  55) - 0.5042) < 5e-4
    assert abs(water_viscosity(273.15 +  60) - 0.4666) < 5e-4
    assert abs(water_viscosity(273.15 +  65) - 0.4334) < 5e-4
    assert abs(water_viscosity(273.15 +  70) - 0.4039) < 5e-4
    assert abs(water_viscosity(273.15 +  75) - 0.3775) < 5e-4
    assert abs(water_viscosity(273.15 +  80) - 0.3538) < 5e-4
    assert abs(water_viscosity(273.15 +  85) - 0.3323) < 5e-4
    assert abs(water_viscosity(273.15 +  90) - 0.3128) < 5e-4
    assert abs(water_viscosity(273.15 +  95) - 0.2949) < 6e-4
    assert abs(water_viscosity(273.15 + 100) - 0.2783) < 2e-3
    warnings.resetwarnings()



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

import pytest

from ..periodic import atomic_number, mass_from_composition, relative_atomic_masses, groups
from ..testing import requires
from ..parsing import formula_to_composition, parsing_library


def test_atomic_number():
    assert atomic_number('U') == 92
    assert atomic_number('carbon') == 6
    assert atomic_number('ununpentium') == 115
    with pytest.raises(ValueError):
        atomic_number('unobtainium')


def test_mass_from_composition():
    mass = mass_from_composition({11: 1, 9: 1})
    assert abs(41.988172443 - mass) < 1e-7


def test_relative_atomic_masses():
    assert relative_atomic_masses[0] == 1.008


def test_groups():
    assert groups[1] == (1, 3, 11, 19, 37, 55, 87)
    assert groups[2] == (4, 12, 20, 38, 56, 88)
    assert groups[13] == (5, 13, 31, 49, 81, 113)
    assert groups[14] == (6, 14, 32, 50, 82, 114)
    assert groups[15] == (7, 15, 33, 51, 83, 115)
    assert groups[16] == (8, 16, 34, 52, 84, 116)
    assert groups[17] == (9, 17, 35, 53, 85, 117)
    assert groups[18] == (2, 10, 18, 36, 54, 86, 118)


@requires(parsing_library)
def test_mass_from_composition__formula():
    mass = mass_from_composition(formula_to_composition('NaF'))
    assert abs(41.988172443 - mass) < 1e-7

    Fminus = mass_from_composition(formula_to_composition('F/-'))
    assert abs(Fminus - 18.998403163 - 5.489e-4) < 1e-7



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

try:
    import numpy as np
except ImportError:
    np = None

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

from .. import ReactionSystem, Equilibrium
from ..units import get_derived_unit, to_unitless, default_units as u


def _dominant_reaction_effects(substance_key, rsys, rates, linthreshy, eqk1, eqk2, eqs):
    tot = np.zeros(rates.shape[0])
    reaction_effects = rsys.per_reaction_effect_on_substance(substance_key)
    data = []
    for ri, n in reaction_effects.items():
        tot += n*rates[..., ri]
        if ri in eqk1:
            otheri = eqk2[eqk1.index(ri)]
            y = n*rates[..., ri] + reaction_effects[otheri]*rates[..., otheri]
            rxn = eqs[eqk1.index(ri)]
        elif ri in eqk2:
            continue
        else:
            y = n*rates[..., ri]
            rxn = rsys.rxns[ri]
        if np.all(np.abs(y) < linthreshy):
            continue
        data.append((y, rxn))
    return data, tot


def _combine_rxns_to_eq(rsys):
    eqk1, eqk2 = zip(*rsys.identify_equilibria())
    eqs = [Equilibrium(
        rsys.rxns[i1].reac,
        rsys.rxns[i1].prod,
        (rsys.rxns[i1].param, rsys.rxns[i2].param),
        inact_reac=rsys.rxns[i1].inact_reac,
        inact_prod=rsys.rxns[i1].inact_prod
    ) for i1, i2 in zip(eqk1, eqk2)]
    return eqk1, eqk2, eqs


def plot_reaction_contributions(
        xyp, rsys, rate_exprs_cb, substance_keys=None, varied=None, axes=None,
        total=False, linthreshy=1e-9, relative=False, xscale='log', yscale='symlog',
        xlabel='Time', ylabel=None, combine_equilibria=False, selection=slice(None),
        unit_registry=None):
    """ Plots per reaction contributions to concentration evolution of a substance.

    Parameters
    ----------
    xyp : ``pyodesys.results.Result`` instance or length 3 tuple or xout,yout,params
    result : pyodesys.results.Result
    substance_key : str

    """
    from pyodesys.results import Result
    if isinstance(xyp, Result):
        xyp = xyp.odesys.to_arrays(xyp.xout, xyp.yout, xyp.params, reshape=False)
    if varied is None:
        varied = xyp[0]
    if xyp[1].shape[-2] != varied.size:
        raise ValueError("Size mismatch between varied and yout")
    if substance_keys is None:
        substance_keys = rsys.substances.keys()
    if axes is None:
        _fig, axes = plt.subplots(len(substance_keys))
    rates = rate_exprs_cb(*xyp)
    if unit_registry is not None:
        time_unit = get_derived_unit(unit_registry, 'time')
        conc_unit = get_derived_unit(unit_registry, 'concentration')
        rates = to_unitless(rates*conc_unit/time_unit, u.molar/u.second)

    eqk1, eqk2, eqs = _combine_rxns_to_eq(rsys) if combine_equilibria else ([], [], [])

    for sk, ax in zip(substance_keys, axes):
        data, tot = _dominant_reaction_effects(sk, rsys, rates, linthreshy, eqk1, eqk2, eqs)
        factor = 1/xyp[1][:, rsys.as_substance_index(sk)] if relative else 1
        if total:
            ax.plot(varied, factor*tot, c='k', label='Total', linewidth=2, ls=':')
        for y, rxn in sorted(data, key=lambda args: args[0][-1], reverse=True):
            ax.plot(varied, factor*y,
                    label=r'$\mathrm{%s}$' % rxn.latex(rsys.substances))

        if rsys.substances[sk].latex_name is None:
            ttl = rsys.substances[sk].name
            ttl_template = '%s'
        else:
            ttl = rsys.substances[sk].latex_name
            ttl_template = r'\mathrm{$%s$}'

        if yscale == 'symlog':
            ax.axhline(linthreshy, linestyle='--', color='k', linewidth=.5)
            ax.axhline(-linthreshy, linestyle='--', color='k', linewidth=.5)
            ax.set_yscale(yscale, linthreshy=linthreshy)
        else:
            ax.set_yscale(yscale)

        if ylabel is None:
            ax.set_ylabel(r'$\frac{d}{dt}\left[%s\right]\ /\ M\cdot s^{-1}$' % ttl)
        else:
            ax.set_ylabel(ylabel)
            ax.set_title(ttl_template % ttl)

        ax.set_xlabel(xlabel)
        ax.set_xscale(xscale)
        ax.legend(loc='best')


def dominant_reactions_graph(concs, rate_exprs_cb, rsys, substance_key, linthreshy=1e-9,
                             fname='dominant_reactions_graph.png', relative=False,
                             combine_equilibria=False, **kwargs):
    from ..util.graph import rsys2graph
    rates = rate_exprs_cb(0, concs)
    eqk1, eqk2, eqs = _combine_rxns_to_eq(rsys) if combine_equilibria else ([], [], [])
    rrate, rxns = zip(*_dominant_reaction_effects(
        substance_key, rsys, rates, linthreshy, eqk1, eqk2, eqs)[0])
    rsys = ReactionSystem(rxns, rsys.substances, rsys.name)
    lg_rrate = np.log10(np.abs(rrate))
    rsys2graph(rsys, fname=fname, penwidths=1 + lg_rrate - np.min(lg_rrate), **kwargs)



from __future__ import division

from chempy.util.testing import requires
from ..integrated import pseudo_irrev, pseudo_rev, binary_irrev, binary_rev

import pytest

try:
    import sympy
except ImportError:
    sympy = None
else:
    one = sympy.S(1)

    t, kf, kb, prod, major, minor = sympy.symbols(
        't kf kb prod major minor', negative=False, nonnegative=True, real=True)

    subsd = {t: one*2, kf: one*3, kb: one*7, major: one*11,
             minor: one*13, prod: one*0}


@requires('sympy')
def test_pseudo_irrev():
    f = pseudo_irrev(t, kf, prod, major, minor, backend=sympy)
    dfdt = f.diff(t)
    num_dfdt = dfdt.subs(subsd)
    assert (num_dfdt - (
        major*kf*(minor - f)
    ).subs(subsd)).simplify() == 0


@requires('sympy')
def test_pseudo_rev():
    f = pseudo_rev(t, kf, kb, prod, major, minor, backend=sympy)
    dfdt = f.diff(t)
    num_dfdt = dfdt.subs(subsd)
    assert (num_dfdt - (major*kf*(minor - f) - kb*f).subs(subsd)).simplify() == 0


@pytest.mark.slow
@requires('sympy')
def test_binary_irrev():
    f = binary_irrev(t, kf, prod, major, minor, backend=sympy)
    dfdt = f.diff(t)
    num_dfdt = dfdt.subs(subsd)
    assert (num_dfdt - (kf*(minor - f)*(major - f)).subs(subsd)).simplify() == 0


@pytest.mark.slow
@requires('sympy')
def test_binary_rev():
    f = binary_rev(t, kf, kb, prod, major, minor, backend=sympy)
    dfdt = f.diff(t)
    num_dfdt = dfdt.subs(subsd)
    ans = kf*(minor - f)*(major - f) - kb*f
    # symbolic susbsitution fails:
    assert abs(float(num_dfdt) - float(ans.subs(subsd))) < 2e-14



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

from ..testing import requires
from ..numutil import broadcast_stack


@requires('numpy')
def test_broadcast_stack():
    import numpy as np
    a = np.arange(9.)
    b = broadcast_stack(np.arange(5.), np.arange(5., 9.))
    assert a.shape == b.shape and np.all(a == b)

    c = broadcast_stack([0, 1, 2], [3], 4, as_scalars=True)
    d = np.array([
        [0, 3, 4],
        [1, 3, 4],
        [2, 3, 4]
    ])
    assert c.shape == d.shape and np.all(c == d)



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

import math

import pytest

from chempy import Reaction
from chempy.units import allclose, Backend, to_unitless, units_library, default_units as u
from chempy.util._expr import Log10, Constant
from chempy.util.testing import requires
from ..rates import MassAction, Radiolytic
from .._rates import TPoly, RTPoly, ShiftedTPoly, ShiftedRTPoly, ShiftedLog10TPoly, TPiecewise


def test_TPolyMassAction():
    r = Reaction({'A': 2, 'B': 1}, {'C': 1}, inact_reac={'B': 1})
    p = MassAction(ShiftedTPoly([273.15, 7, .2, .03, .004]))
    res = p({'A': 11, 'B': 13, 'temperature': 298.15}, reaction=r)
    ref = 7 + .2*25 + 0.03 * 25**2 + 0.004 * 25**3
    assert abs(res - ref*13*11**2) < 1e-15


def test_ShiftedTPoly_MassAction():
    rate = MassAction(ShiftedTPoly([100, 2, 5, 7]))
    assert rate.args[0].args == [100, 2, 5, 7]
    r = Reaction({'A': 2, 'B': 1}, {'P': 1}, rate)
    res = r.rate_expr()({'A': 11, 'B': 13, 'temperature': 273.15}, reaction=r)
    x = 273.15-100
    ref = 11*11*13*(2 + 5*x + 7*x**2)
    assert abs((res - ref)/ref) < 1e-14


@requires(units_library)
def test_ShiftedTPoly__units():
    stp1 = ShiftedTPoly([273.15*u.kelvin, 5, 7/u.kelvin])
    allclose(stp1({'temperature': 274.15*u.kelvin}), 5+7)
    allclose(stp1({'temperature': 273.15*u.kelvin}), 5)
    stp2 = ShiftedTPoly([273.15*u.kelvin, 5*u.m, 7*u.m/u.kelvin, 13*u.m*u.kelvin**-2])
    allclose(stp2({'temperature': 274.15*u.kelvin}), (5+7+13)*u.m)
    allclose(stp2({'temperature': 273.15*u.kelvin}), 5*u.m)


@requires(units_library)
def test_TPolyMassAction__units():
    Mps = u.molar/u.second
    kunit = 1/u.molar**2/u.second
    r = Reaction({'A': 2, 'B': 1}, {'C': 1}, None, {'B': 1})
    p = MassAction(ShiftedTPoly([273.15*u.K, 7*kunit, .2*kunit/u.K, .03*kunit/u.K**2, .004*kunit/u.K**3]))
    res = p({'A': 11*u.molar, 'B': 13*u.molar, 'temperature': 298.15*u.K}, reaction=r)
    ref = 7 + .2*25 + 0.03 * 25**2 + 0.004 * 25**3
    assert abs(res - ref*13*11**2*Mps) < 1e-15


def test_RTPolyMassAction():
    p = MassAction(ShiftedRTPoly([273.15, 7, .2, .03, .004]))
    r = Reaction({'A': 2, 'B': 1}, {'C': 1}, inact_reac={'B': 1})
    res = p({'A': 11, 'B': 13, 'temperature': 298.15}, reaction=r)
    ref = 7 + .2/25 + 0.03 / 25**2 + 0.004 / 25**3
    assert abs(res - ref*13*11**2) < 1e-15


@requires(units_library)
def test_RTPolyMassAction__units():
    Mps = u.molar/u.second
    kunit = 1/u.molar**2/u.second
    p = MassAction(ShiftedRTPoly([273.15*u.K, 7*kunit, .2*kunit*u.K, .03*kunit*u.K**2, .004*kunit*u.K**3]))
    r = Reaction({'A': 2, 'B': 1}, {'C': 1}, None, {'B': 1})
    res = p({'A': 11*u.molar, 'B': 13*u.molar, 'temperature': 298.15*u.K}, reaction=r)
    ref = 7 + .2/25 + 0.03 / 25**2 + 0.004 / 25**3
    assert abs(res - ref*13*11**2*Mps) < 1e-15


def test_Log10TPolyMassAction():
    p = MassAction(10**ShiftedTPoly([273.15, .7, .02, .003, .0004]))
    r = Reaction({'A': 2, 'B': 1}, {'C': 1}, p, {'B': 1})
    res = p({'A': 11, 'B': 13, 'temperature': 298.15}, reaction=r)
    ref = 10**(.7 + .02*25 + 0.003 * 25**2 + 0.0004 * 25**3)
    assert abs(res - ref*13*11**2) < 1e-15


@requires(units_library)
def test_Log10TPolyMassAction__units():
    Mps = u.molar/u.second
    kunit = 1/u.molar**2/u.second
    p = MassAction(Constant(kunit)*10**ShiftedTPoly([273.15*u.K, .7, .02/u.K, .003/u.K**2, .0004/u.K**3]))
    r = Reaction({'A': 2, 'B': 1}, {'C': 1}, p, {'B': 1})
    res = p({'A': 11*u.molar, 'B': 13*u.molar, 'temperature': 298.15*u.K}, reaction=r)
    ref = 10**(.7 + .02*25 + 0.003 * 25**2 + 0.0004 * 25**3)
    assert abs(res - ref*13*11**2*Mps) < 1e-15


def test_TPolyInLog10MassAction():
    p = MassAction(ShiftedLog10TPoly([2, 0.3, .2, .03, .004]))
    r = Reaction({'A': 2, 'B': 1}, {'C': 1}, p, {'B': 1})
    lgT = Log10('temperature')
    lgTref = Log10('Tref')
    res = p({'A': 11, 'B': 13, 'temperature': 298.15,
             'log10_temperature': lgT, 'log10_Tref': lgTref}, reaction=r)
    _T = math.log10(298.15) - 2
    ref = .3 + .2*_T + 0.03 * _T**2 + 0.004 * _T**3
    assert abs(res - ref*13*11**2) < 1e-15


@requires(units_library)
def test_TPolyInLog10MassAction__units():
    Mps = u.molar/u.second
    kunit = 1/u.molar**2/u.second
    p = MassAction(Constant(kunit)*ShiftedLog10TPoly([2, 0.3, .2, .03, .004]))
    lgT = Log10('temperature'/Constant(u.K))
    r = Reaction({'A': 2, 'B': 1}, {'C': 1}, p, {'B': 1})
    res = p({'A': 11*u.molar, 'B': 13*u.molar, 'temperature': 298.15*u.K,
             'log10_temperature': lgT}, backend=Backend(), reaction=r)
    _T = math.log10(298.15) - 2
    ref = (.3 + .2*_T + 0.03 * _T**2 + 0.004 * _T**3) * 13 * 11**2 * Mps
    assert abs(res - ref) < 1e-15*Mps


def test_TPolyRadiolytic():
    tpr = Radiolytic(ShiftedTPoly([273.15, 1.85e-7, 1e-9]))
    res = tpr({'doserate': 0.15, 'density': 0.998, 'temperature': 298.15})
    assert abs(res - 0.15*0.998*2.1e-7) < 1e-15


@requires(units_library)
def test_TPolyRadiolytic__units():

    def _check(r):
        res = r({'doserate': 0.15*u.gray/u.second,
                 'density': 0.998*u.kg/u.decimetre**3,
                 'temperature': 298.15*u.K})
        ref = 0.15*0.998*2.1e-7*u.molar/u.second
        assert abs(to_unitless((res - ref)/ref)) < 1e-15

    _check(Radiolytic(ShiftedTPoly([273.15*u.K, 1.85e-7*u.mol/u.joule, 1e-9*u.mol/u.joule/u.K])))


def test_TPiecewisePolyMassAction():
    tp1 = TPoly([10, 0.1])
    tp2 = ShiftedTPoly([273.15, 37.315, -0.1])
    pwp = MassAction(TPiecewise([0, tp1, 273.15, tp2, 373.15]))
    r = Reaction({'A': 2, 'B': 1}, {'C': 1}, inact_reac={'B': 1})
    res1 = pwp({'A': 11, 'B': 13, 'temperature': 198.15}, reaction=r)
    ref1 = 11*11*13*29.815
    assert abs((res1-ref1)/ref1) < 1e-14
    res2 = pwp({'A': 11, 'B': 13, 'temperature': 298.15}, reaction=r)
    ref2 = 11*11*13*(37.315 - 25*0.1)
    assert abs((res2-ref2)/ref2) < 1e-14
    with pytest.raises(ValueError):
        pwp({'A': 11, 'B': 13, 'temperature': 398.15}, reaction=r)


@requires(units_library)
def test_PiecewiseTPolyMassAction__units():
    tp1 = TPoly([10/u.molar**2/u.second, 0.1/u.molar**2/u.second/u.kelvin])
    tp2 = ShiftedTPoly([273.15*u.kelvin, 37.315/u.molar**2/u.second, -0.1/u.molar**2/u.second/u.kelvin])
    pwp = MassAction(TPiecewise([0*u.kelvin, tp1, 273.15*u.kelvin, tp2, float('inf')*u.kelvin]))
    r = Reaction({'A': 2, 'B': 1}, {'C': 1}, pwp, {'B': 1})
    res1 = pwp({'A': 11*u.molar, 'B': 13*u.molar, 'temperature': 198.15*u.kelvin}, reaction=r)
    ref1 = 11*11*13*29.815 * u.molar/u.second
    assert abs((res1-ref1)/ref1) < 1e-14


@requires('sympy')
def test_PiecewiseTPolyMassAction__sympy():
    import sympy as sp
    tp1 = TPoly([10, 0.1])
    tp2 = ShiftedTPoly([273.15, 37.315, -0.1])
    pwp = MassAction(TPiecewise([0, tp1, 273.15, tp2, 373.15]))
    T = sp.Symbol('T')
    r = Reaction({'A': 2, 'B': 1}, {'C': 1}, inact_reac={'B': 1})
    res1 = pwp({'A': 11, 'B': 13, 'temperature': T}, backend=sp, reaction=r)
    ref1 = 11**2 * 13 * sp.Piecewise(
        (10+0.1*T, sp.And(0 <= T, T <= 273.15)),
        (37.315 - 0.1*(T-273.15), sp.And(273.15 <= T, T <= 373.15)),
        (sp.Symbol('NAN'), True)
    )
    assert res1 == ref1


def test_Log10PiecewiseRTPolyMassAction():
    p1 = RTPoly([12.281, -3.768e2, -6.673e4, -1.075e7])
    p2 = RTPoly([-47.532, 4.92, -1.036, 0.0])
    ratex = MassAction(10**TPiecewise([293.15, p1, 423.15, p2, 623.15]))
    r = Reaction({'e-(aq)': 2}, {'H2': 1, 'OH-': 2}, ratex, {'H2O': 2})
    res = ratex({'e-(aq)': 1e-13, 'temperature': 293.15}, reaction=r)
    ref = 6.20e9*1e-26
    assert abs((res-ref)/ref) < 6e-3


@requires(units_library)
def test_TPiecewise():
    expr0 = ShiftedTPoly([273.15*u.K, 10/u.molar/u.s, 0.1/u.molar/u.s/u.K])
    expr1 = ShiftedTPoly([298.15*u.K, 12.5/u.molar/u.s, 0/u.molar/u.s/u.K, 2/u.molar/u.s/u.K**2])
    pwma = MassAction(TPiecewise([273.15*u.K, expr0, 298.15*u.K, expr1, 373.15*u.K]))
    r = Reaction({'e-(aq)': 2}, {'H2': 1, 'OH-': 2}, inact_reac={'H2O': 2})
    res0 = pwma({'temperature': 293.15*u.K, 'e-(aq)': 1e-13*u.molar}, reaction=r)
    ref0 = 12*1e-26 * u.molar/u.s
    assert allclose(res0, ref0)
    assert not allclose(res0, 2*ref0)

    res1 = pwma({'temperature': 300.15*u.K, 'e-(aq)': 2e-13*u.molar}, reaction=r)
    ref1 = 20.5*4e-26 * u.molar/u.s
    assert allclose(res1, ref1)
    assert not allclose(res1, ref1/2)


@requires(units_library)
def test_MassAction__rate_coeff():
    perMolar_perSecond = u.perMolar_perSecond

    p1 = MassAction(Constant(perMolar_perSecond) * 10**RTPoly([1, 2*u.kelvin, 3*u.kelvin**2]))
    rcoeff1 = p1.rate_coeff({'temperature': 283.15*u.K})
    ref1 = 10**(1 + 2/283.15 + 3/283.15**2) * perMolar_perSecond
    assert allclose(rcoeff1, ref1)
    rxn1 = Reaction({'A', 'B'}, {'C'}, p1)
    rat1 = rxn1.rate({'A': 2, 'B': 3, 'temperature': 283.15*u.K})
    assert allclose(rat1['A'], -2*3*ref1)
    assert allclose(rat1['B'], -2*3*ref1)
    assert allclose(rat1['C'], 2*3*ref1)



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

from functools import reduce
from operator import mul, add

try:
    import numpy as np
    from numpy import any as _any

    def prodpow(bases, exponents):
        """
        Examples
        --------
        >>> prodpow([2, 3], np.array([[0, 1], [1, 2]]))
        array([ 3, 18])

        """
        exponents = np.asarray(exponents)
        return np.multiply.reduce(bases**exponents, axis=-1)

except ImportError:  # no NumPy available
    def _any(arg):
        if arg is True:
            return True
        if arg is False:
            return False
        return any(arg)

    def prodpow(bases, exponents):
        """
        Examples
        --------
        >>> prodpow([2, 3], [[0, 1], [1, 2]])
        [3, 18]

        """
        result = []
        for row in exponents:
            res = 1
            for b, e in zip(bases, row):
                res *= b**e
            result.append(res)
        return result


def get_backend(backend):
    if isinstance(backend, str):
        backend = __import__(backend)
    if backend is None:
        try:
            import numpy as backend
        except ImportError:
            import math as backend
    return backend


def intdiv(p, q):
    """ Integer divsions which rounds toward zero

    Examples
    --------
    >>> intdiv(3, 2)
    1
    >>> intdiv(-3, 2)
    -1
    >>> -3 // 2
    -2

    """
    r = p // q
    if r < 0 and q*r != p:
        r += 1
    return r


def reducemap(args, reduce_op, map_op):
    return reduce(reduce_op, map(map_op, *args))


def vec_dot_vec(vec1, vec2):
    # return np.dot(vec1, vec2)
    # return np.add.reduce(np.multiply(vec1, vec2))
    return reducemap((vec1, vec2), add, mul)


def mat_dot_vec(iter_mat, iter_vec, iter_term=None):  # pure python (slow)
    if iter_term is None:
        return [vec_dot_vec(row, iter_vec) for row in iter_mat]
    else:
        # daxpy
        return [vec_dot_vec(row, iter_vec) + term for row, term
                in zip(iter_mat, iter_term)]


# def composition_balance(substances, concs, composition_number):
#     if not hasattr(concs, 'ndim') or concs.ndim == 1:
#         res = 0
#     elif concs.ndim == 2:
#         res = np.zeros(concs.shape[0])
#         concs = concs.T
#     else:
#         raise NotImplementedError
#     for s, c in zip(substances, concs):
#         res += s.composition.get(composition_number, 0)*c
#     return res



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

from .util.pyutil import ChemPyDeprecationWarning

import warnings

from .electrolytes import (
    A, B, limiting_log_gamma, extended_log_gamma, davies_log_gamma,
    limiting_activity_product, extended_activity_product,
    davies_activity_product
)


warnings.warn("use .electrolytes instead of .debye_huckel", ChemPyDeprecationWarning)



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

import pytest

from ..parsing import (
    formula_to_composition, formula_to_html, formula_to_latex, formula_to_unicode,
    parsing_library, to_reaction
)
from ..testing import requires


@requires(parsing_library)
def test_formula_to_composition():
    assert formula_to_composition('H2O') == {1: 2, 8: 1}
    assert formula_to_composition('Fe/3+') == {0: 3, 26: 1}
    assert formula_to_composition('Fe+3') == {0: 3, 26: 1}
    assert formula_to_composition('Na/+') == {0: 1, 11: 1}
    assert formula_to_composition('Na+1') == {0: 1, 11: 1}
    assert formula_to_composition('Na+') == {0: 1, 11: 1}
    assert formula_to_composition('Cl/-') == {0: -1, 17: 1}
    assert formula_to_composition('Cl-') == {0: -1, 17: 1}
    assert formula_to_composition('NaCl') == {11: 1, 17: 1}
    assert formula_to_composition('NaCl(s)') == {11: 1, 17: 1}
    assert formula_to_composition('Fe(SCN)2/+') == {
        0: 1, 6: 2, 7: 2, 16: 2, 26: 1}
    assert formula_to_composition('Fe(SCN)2+') == {
        0: 1, 6: 2, 7: 2, 16: 2, 26: 1}
    assert formula_to_composition('Fe(SCN)2+1') == {
        0: 1, 6: 2, 7: 2, 16: 2, 26: 1}
    assert formula_to_composition('((H2O)2OH)12') == {1: 60, 8: 36}

    # Special case: solvated electron:
    assert formula_to_composition('e-') == {0: -1}
    assert formula_to_composition('e-1') == {0: -1}
    assert formula_to_composition('e-(aq)') == {0: -1}
    assert formula_to_composition('SO4-2(aq)') == {0: -2, 8: 4, 16: 1}

    # prefixes and suffixes
    assert formula_to_composition('.NO2(g)') == {7: 1, 8: 2}
    assert formula_to_composition('.NH2') == {1: 2, 7: 1}
    assert formula_to_composition('ONOOH') == {1: 1, 7: 1, 8: 3}
    assert formula_to_composition('.ONOO') == {7: 1, 8: 3}
    assert formula_to_composition('.NO3/2-') == {0: -2, 7: 1, 8: 3}
    assert formula_to_composition('.NO3-2') == {0: -2, 7: 1, 8: 3}

    with pytest.raises(ValueError):
        formula_to_composition('F-F')

    assert formula_to_composition('alpha-FeOOH(s)') == {1: 1, 8: 2, 26: 1}
    assert formula_to_composition('epsilon-Zn(OH)2(s)') == {1: 2, 8: 2, 30: 1}

    # crystal water
    assert formula_to_composition('Na2CO3.7H2O(s)') == {11: 2, 6: 1, 8: 10, 1: 14}


@requires(parsing_library)
def test_to_reaction():
    from chempy.chemistry import Reaction, Equilibrium
    rxn = to_reaction(
        "H+ + OH- -> H2O; 1.4e11; ref={'doi': '10.1039/FT9908601539'}",
        'H+ OH- H2O'.split(), '->', Reaction)
    assert rxn.__class__ == Reaction

    assert rxn.reac['H+'] == 1
    assert rxn.reac['OH-'] == 1
    assert rxn.prod['H2O'] == 1
    assert rxn.param == 1.4e11
    assert rxn.ref['doi'].startswith('10.')

    eq = to_reaction("H+ + OH- = H2O; 1e-14; ref='rt, [H2O] == 1 M'",
                     'H+ OH- H2O'.split(), '=', Equilibrium)
    assert eq.__class__ == Equilibrium

    assert eq.reac['H+'] == 1
    assert eq.reac['OH-'] == 1
    assert eq.prod['H2O'] == 1
    assert eq.ref.startswith('rt')

    for s in ['2 e-(aq) + (2 H2O) -> H2 + 2 OH- ; 1e6 ; ',
              '2 * e-(aq) + (2 H2O) -> 1 * H2 + 2 * OH- ; 1e6 ; ']:
        rxn2 = to_reaction(s, 'e-(aq) H2 OH- H2O'.split(), '->', Reaction)
        assert rxn2.__class__ == Reaction
        assert rxn2.reac['e-(aq)'] == 2
        assert rxn2.inact_reac['H2O'] == 2
        assert rxn2.prod['H2'] == 1
        assert rxn2.prod['OH-'] == 2
        assert rxn2.param == 1e6

    r1 = to_reaction("-> H2O", None, '->', Reaction)
    assert r1.reac == {}
    assert r1.prod == {'H2O': 1}
    assert r1.param is None

    r2 = to_reaction("H2O ->", None, '->', Reaction)
    assert r2.reac == {'H2O': 1}
    assert r2.prod == {}
    assert r2.param is None

    from chempy.kinetics.rates import MassAction
    ma = MassAction([3.14])
    r3 = to_reaction("H+ + OH- -> H2O", None, '->', Reaction, param=ma)
    assert r3.param.args == [3.14]

    rxn3 = to_reaction("H2O + H2O -> H3O+ + OH-", 'H3O+ OH- H2O'.split(), '->', Reaction)
    assert rxn3.reac == {'H2O': 2} and rxn3.prod == {'H3O+': 1, 'OH-': 1}

    rxn4 = to_reaction("2 e-(aq) + (2 H2O) + (2 H+) -> H2 + 2 H2O", 'e-(aq) H2 H2O H+'.split(), '->', Reaction)
    assert rxn4.reac == {'e-(aq)': 2} and rxn4.inact_reac == {'H2O': 2, 'H+': 2} and rxn4.prod == {'H2': 1, 'H2O': 2}


@requires(parsing_library)
def test_formula_to_latex():
    assert formula_to_latex('H2O') == 'H_{2}O'
    assert formula_to_latex('C6H6/+') == 'C_{6}H_{6}^{+}'
    assert formula_to_latex('Fe(CN)6/3-') == 'Fe(CN)_{6}^{3-}'
    assert formula_to_latex('Fe(CN)6-3') == 'Fe(CN)_{6}^{3-}'
    assert formula_to_latex('C18H38/2+') == 'C_{18}H_{38}^{2+}'
    assert formula_to_latex('C18H38/+2') == 'C_{18}H_{38}^{2+}'
    assert formula_to_latex('C18H38+2') == 'C_{18}H_{38}^{2+}'
    assert formula_to_latex('((H2O)2OH)12') == '((H_{2}O)_{2}OH)_{12}'
    assert formula_to_latex('NaCl') == 'NaCl'
    assert formula_to_latex('NaCl(s)') == 'NaCl(s)'
    assert formula_to_latex('e-(aq)') == 'e^{-}(aq)'
    assert formula_to_latex('Ca+2(aq)') == 'Ca^{2+}(aq)'
    assert formula_to_latex('.NO2(g)') == r'^\bullet NO_{2}(g)'
    assert formula_to_latex('.NH2') == r'^\bullet NH_{2}'
    assert formula_to_latex('ONOOH') == 'ONOOH'
    assert formula_to_latex('.ONOO') == r'^\bullet ONOO'
    assert formula_to_latex('.NO3/2-') == r'^\bullet NO_{3}^{2-}'
    assert formula_to_latex('.NO3-2') == r'^\bullet NO_{3}^{2-}'
    assert formula_to_latex('alpha-FeOOH(s)') == r'\alpha-FeOOH(s)'
    assert formula_to_latex('epsilon-Zn(OH)2(s)') == (
        r'\varepsilon-Zn(OH)_{2}(s)')
    assert formula_to_latex('Na2CO3.7H2O(s)') == 'Na_{2}CO_{3}\cdot 7H_{2}O(s)'
    assert formula_to_latex('Na2CO3.1H2O(s)') == 'Na_{2}CO_{3}\cdot H_{2}O(s)'


@requires(parsing_library)
def test_formula_to_unicoce():
    assert formula_to_unicode('NH4+') == u'NH₄⁺'
    assert formula_to_unicode('H2O') == u'H₂O'
    assert formula_to_unicode('C6H6/+') == u'C₆H₆⁺'
    assert formula_to_unicode('Fe(CN)6/3-') == u'Fe(CN)₆³⁻'
    assert formula_to_unicode('Fe(CN)6-3') == u'Fe(CN)₆³⁻'
    assert formula_to_unicode('C18H38/2+') == u'C₁₈H₃₈²⁺'
    assert formula_to_unicode('C18H38/+2') == u'C₁₈H₃₈²⁺'
    assert formula_to_unicode('C18H38+2') == u'C₁₈H₃₈²⁺'
    assert formula_to_unicode('((H2O)2OH)12') == u'((H₂O)₂OH)₁₂'
    assert formula_to_unicode('NaCl') == u'NaCl'
    assert formula_to_unicode('NaCl(s)') == u'NaCl(s)'
    assert formula_to_unicode('e-(aq)') == u'e⁻(aq)'
    assert formula_to_unicode('Ca+2(aq)') == u'Ca²⁺(aq)'
    assert formula_to_unicode('.NO2(g)') == u'⋅NO₂(g)'
    assert formula_to_unicode('.NH2') == u'⋅NH₂'
    assert formula_to_unicode('ONOOH') == u'ONOOH'
    assert formula_to_unicode('.ONOO') == u'⋅ONOO'
    assert formula_to_unicode('.NO3/2-') == u'⋅NO₃²⁻'
    assert formula_to_unicode('.NO3-2') == u'⋅NO₃²⁻'
    assert formula_to_unicode('alpha-FeOOH(s)') == u'α-FeOOH(s)'
    assert formula_to_unicode('epsilon-Zn(OH)2(s)') == u'ε-Zn(OH)₂(s)'
    assert formula_to_unicode('Na2CO3.7H2O(s)') == u'Na₂CO₃·7H₂O(s)'
    assert formula_to_unicode('Na2CO3.1H2O(s)') == u'Na₂CO₃·H₂O(s)'


@requires(parsing_library)
def test_formula_to_html():
    assert formula_to_html('H2O') == 'H<sub>2</sub>O'
    assert formula_to_html('C6H6/+') == 'C<sub>6</sub>H<sub>6</sub><sup>+</sup>'
    assert formula_to_html('Fe(CN)6/3-') == 'Fe(CN)<sub>6</sub><sup>3-</sup>'
    assert formula_to_html('Fe(CN)6-3') == 'Fe(CN)<sub>6</sub><sup>3-</sup>'
    assert formula_to_html('C18H38/2+') == 'C<sub>18</sub>H<sub>38</sub><sup>2+</sup>'
    assert formula_to_html('C18H38/+2') == 'C<sub>18</sub>H<sub>38</sub><sup>2+</sup>'
    assert formula_to_html('C18H38+2') == 'C<sub>18</sub>H<sub>38</sub><sup>2+</sup>'
    assert formula_to_html('((H2O)2OH)12') == '((H<sub>2</sub>O)<sub>2</sub>OH)<sub>12</sub>'
    assert formula_to_html('NaCl') == 'NaCl'
    assert formula_to_html('NaCl(s)') == 'NaCl(s)'
    assert formula_to_html('e-(aq)') == 'e<sup>-</sup>(aq)'
    assert formula_to_html('Ca+2(aq)') == 'Ca<sup>2+</sup>(aq)'
    assert formula_to_html('.NO2(g)') == r'&sdot;NO<sub>2</sub>(g)'
    assert formula_to_html('.NH2') == r'&sdot;NH<sub>2</sub>'
    assert formula_to_html('ONOOH') == 'ONOOH'
    assert formula_to_html('.ONOO') == r'&sdot;ONOO'
    assert formula_to_html('.NO3/2-') == r'&sdot;NO<sub>3</sub><sup>2-</sup>'
    assert formula_to_html('.NO3-2') == r'&sdot;NO<sub>3</sub><sup>2-</sup>'
    assert formula_to_html('alpha-FeOOH(s)') == r'&alpha;-FeOOH(s)'
    assert formula_to_html('epsilon-Zn(OH)2(s)') == (
        r'&epsilon;-Zn(OH)<sub>2</sub>(s)')
    assert formula_to_html('Na2CO3.7H2O(s)') == 'Na<sub>2</sub>CO<sub>3</sub>&sdot;7H<sub>2</sub>O(s)'
    assert formula_to_html('Na2CO3.1H2O(s)') == 'Na<sub>2</sub>CO<sub>3</sub>&sdot;H<sub>2</sub>O(s)'



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

from itertools import chain

import pytest

from ..util.testing import requires
from ..util.parsing import parsing_library
from ..units import default_units, units_library
from ..chemistry import Substance, Reaction
from ..reactionsystem import ReactionSystem


@requires(parsing_library, 'numpy')
def test_ReactionSystem():
    import numpy as np
    kw = dict(substance_factory=Substance.from_formula)
    r1 = Reaction.from_string('H2O -> H+ + OH-', 'H2O H+ OH-', name='r1')
    rs = ReactionSystem([r1], 'H2O H+ OH-', **kw)
    r2 = Reaction.from_string('H2O -> 2 H+ + OH-', 'H2O H+ OH-', name='r2')
    with pytest.raises(ValueError):
        ReactionSystem([r2], 'H2O H+ OH-', **kw)
    with pytest.raises(ValueError):
        ReactionSystem([r1, r1], 'H2O H+ OH-', **kw)
    assert rs.as_substance_index('H2O') == 0
    assert rs.as_substance_index(0) == 0
    varied, varied_keys = rs.per_substance_varied({'H2O': 55.4, 'H+': 1e-7, 'OH-': 1e-7},
                                                  {'H+': [1e-8, 1e-9, 1e-10, 1e-11], 'OH-': [1e-3, 1e-2]})
    assert varied_keys == ('H+', 'OH-')
    assert len(varied.shape) == 3
    assert varied.shape[:-1] == (4, 2)
    assert varied.shape[-1] == 3
    assert np.all(varied[..., 0] == 55.4)
    assert np.all(varied[:, 1, 2] == 1e-2)

    assert rs['r1'] is r1
    rs.rxns.append(r2)
    assert rs['r2'] is r2
    with pytest.raises(KeyError):
        rs['r3']
    rs.rxns.append(Reaction({}, {}, 0, name='r2', checks=()))
    with pytest.raises(ValueError):
        rs['r2']


@requires(parsing_library)
def test_ReactionSystem__missing_substances_from_keys():
    r1 = Reaction({'H2O'}, {'H+', 'OH-'})
    with pytest.raises(ValueError):
        ReactionSystem([r1], substances={'H2O': Substance.from_formula('H2O')})
    kw = dict(missing_substances_from_keys=True, substance_factory=Substance.from_formula)
    rs = ReactionSystem([r1], substances={'H2O': Substance.from_formula('H2O')}, **kw)
    assert rs.substances['OH-'].composition == {0: -1, 1: 1, 8: 1}


@requires(parsing_library)
def test_ReactionSystem__check_balance():
    rs1 = ReactionSystem.from_string('\n'.join(['2 NH3 -> N2 + 3 H2', 'N2H4 -> N2 + 2 H2']))
    assert rs1.check_balance(strict=True)
    rs2 = ReactionSystem.from_string('\n'.join(['2 A -> B', 'B -> 2A']),
                                     substance_factory=Substance)
    assert not rs2.check_balance(strict=True)
    assert rs2.composition_balance_vectors() == ([], [])


def test_ReactionSystem__per_reaction_effect_on_substance():
    rs = ReactionSystem([Reaction({'H2': 2, 'O2': 1}, {'H2O': 2})])
    assert rs.per_reaction_effect_on_substance('H2') == {0: -2}
    assert rs.per_reaction_effect_on_substance('O2') == {0: -1}
    assert rs.per_reaction_effect_on_substance('H2O') == {0: 2}


def test_ReactionSystem__rates():
    rs = ReactionSystem([Reaction({'H2O'}, {'H+', 'OH-'}, 11)])
    assert rs.rates({'H2O': 3, 'H+': 5, 'OH-': 7}) == {'H2O': -11*3, 'H+': 11*3, 'OH-': 11*3}


def test_ReactionSystem__rates__cstr():
    k = 11
    rs = ReactionSystem([Reaction({'H2O2': 2}, {'O2': 1, 'H2O': 2}, k)])
    c0 = {'H2O2': 3, 'O2': 5, 'H2O': 53}
    fr = 7
    fc = {'H2O2': 13, 'O2': 17, 'H2O': 23}
    r = k*c0['H2O2']**2
    ref = {
        'H2O2': -2*r + fr*fc['H2O2'] - fr*c0['H2O2'],
        'O2': r + fr*fc['O2'] - fr*c0['O2'],
        'H2O': 2*r + fr*fc['H2O'] - fr*c0['H2O']
    }
    variables = dict(chain(c0.items(), [('fc_'+key, val) for key, val in fc.items()], [('fr', fr)]))
    for fck in (['fc_'+key for key in rs.substances], 'fc_'):
        assert rs.rates(variables, cstr_fr_fc=('fr', fck)) == ref


@requires('numpy')
def test_ReactionSystem__html_tables():
    r1 = Reaction({'A': 2}, {'A'}, name='R1')
    r2 = Reaction({'A'}, {'A': 2}, name='R2')
    rs = ReactionSystem([r1, r2])
    ut, unc = rs.unimolecular_html_table()
    assert unc == {0}
    from chempy.printing import html
    assert html(ut) == u'<table><tr><td>A</td><td ><a title="1: A → 2 A">R2</a></td></tr></table>'

    bt, bnc = rs.bimolecular_html_table()
    assert bnc == {1}
    assert html(bt) == u'<table><th></th><th>A</th>\n<tr><td>A</td><td ><a title="0: 2 A → A">R1</a></td></tr></table>'


@requires(parsing_library, 'numpy')
def test_ReactionSystem__substance_factory():
    r1 = Reaction.from_string('H2O -> H+ + OH-', 'H2O H+ OH-')
    rs = ReactionSystem([r1], 'H2O H+ OH-',
                        substance_factory=Substance.from_formula)
    assert rs.net_stoichs(['H2O']) == [-1]
    assert rs.net_stoichs(['H+']) == [1]
    assert rs.net_stoichs(['OH-']) == [1]
    assert rs.substances['H2O'].composition[8] == 1
    assert rs.substances['OH-'].composition[0] == -1
    assert rs.substances['H+'].charge == 1


@requires(units_library)
def test_ReactionSystem__as_per_substance_array_dict():
    mol = default_units.mol
    m = default_units.metre
    M = default_units.molar
    rs = ReactionSystem([], [Substance('H2O')])
    c = rs.as_per_substance_array({'H2O': 1*M}, unit=M)
    assert c.dimensionality == M.dimensionality
    assert abs(c[0]/(1000*mol/m**3) - 1) < 1e-16

    c = rs.as_per_substance_array({'H2O': 1})
    with pytest.raises(KeyError):
        c = rs.as_per_substance_array({'H': 1})

    assert rs.as_per_substance_dict([42]) == {'H2O': 42}


@requires(parsing_library)
def test_ReactionSystem__add():
    rs1 = ReactionSystem.from_string('\n'.join(['2 H2O2 -> O2 + 2 H2O', 'H2 + O2 -> H2O2']))
    rs2 = ReactionSystem.from_string('\n'.join(['2 NH3 -> N2 + 3 H2']))
    rs3 = rs1 + rs2
    assert rs1 == rs1
    assert rs1 != rs2
    assert rs3 != rs1
    assert len(rs1.rxns) == 2 and len(rs2.rxns) == 1 and len(rs3.rxns) == 3
    for k in 'H2O2 O2 H2O H2 NH3 N2'.split():
        assert k in rs3.substances
    rs1 += rs2
    assert len(rs1.rxns) == 3 and len(rs2.rxns) == 1
    assert rs1 == rs3

    rs4 = ReactionSystem.from_string("H2O -> H+ + OH-; 1e-4")
    rs4 += [Reaction({'H+', 'OH-'}, {'H2O'}, 1e10)]
    res = rs4.rates({'H2O': 1, 'H+': 1e-7, 'OH-': 1e-7})
    for k in 'H2O H+ OH-'.split():
        assert abs(res[k]) < 1e-16

    rs5 = ReactionSystem.from_string("H3O+ -> H+ + H2O")
    rs6 = rs4 + rs5
    rs7 = rs6 + (Reaction.from_string("H+ + H2O -> H3O+"),)
    assert len(rs7.rxns) == 4


@requires(parsing_library)
def test_ReactionSystem__from_string():
    rs = ReactionSystem.from_string('-> H + OH; Radiolytic(2.1e-7)', checks=())
    assert rs.rxns[0].reac == {}
    assert rs.rxns[0].prod == {'H': 1, 'OH': 1}
    assert rs.rxns[0].param.args == [2.1e-7]
    ref = 2.1e-7 * 0.15 * 998
    assert rs.rates({'doserate': .15, 'density': 998}) == {'H': ref, 'OH': ref}

    r2, = ReactionSystem.from_string("H2O + H2O + H+ -> H3O+ + H2O").rxns
    assert r2.reac == {'H2O': 2, 'H+': 1}
    assert r2.prod == {'H2O': 1, 'H3O+': 1}

    r3, = ReactionSystem.from_string('(H2O) -> e-(aq) + H+ + OH; Radiolytic(2.1e-7*mol/J)').rxns
    assert len(r3.reac) == 0 and r3.inact_reac == {'H2O': 1}
    assert r3.prod == {'e-(aq)': 1, 'H+': 1, 'OH': 1}
    from chempy.kinetics.rates import Radiolytic
    mol, J = default_units.mol, default_units.J
    assert r3.param == Radiolytic(2.1e-7*mol/J)
    assert r3.param != Radiolytic(2.0e-7*mol/J)
    assert r3.param != Radiolytic(2.1e-7)
    assert r3.order() == 0


@requires(parsing_library)
def test_ReactionSystem__from_string__string_rate_const():
    rsys = ReactionSystem.from_string("H+ + OH- -> H2O; 'kf'")
    r2, = rsys.rxns
    assert r2.reac == {'OH-': 1, 'H+': 1}
    assert r2.prod == {'H2O': 1}
    r2str = r2.string(rsys.substances, with_param=True)
    assert r2str.endswith('; kf')


@requires('numpy')
def test_ReactionSystem__upper_conc_bounds():
    rs = ReactionSystem.from_string('\n'.join(['2 NH3 -> N2 + 3 H2', 'N2H4 -> N2 +   2  H2']))
    c0 = {'NH3': 5, 'N2': 7, 'H2': 11, 'N2H4': 2}
    _N = 5 + 14 + 4
    _H = 15 + 22 + 8
    ref = {
        'NH3': min(_N, _H/3),
        'N2': _N/2,
        'H2': _H/2,
        'N2H4': min(_N/2, _H/4),
    }
    res = rs.as_per_substance_dict(rs.upper_conc_bounds(c0))
    assert res == ref


@requires('numpy')
def test_ReactionSystem__upper_conc_bounds__a_substance_no_composition():
    rs = ReactionSystem.from_string("""
    H2O -> e-(aq) + H2O+
    H2O+ + e-(aq) -> H2O
    """)
    c0 = {'H2O': 55.0, 'e-(aq)': 2e-3, 'H2O+': 3e-3}
    _O = 55 + 3e-3
    _H = 2*55 + 2*3e-3
    ref = {
        'H2O': min(_O, _H/2),
        'e-(aq)': float('inf'),
        'H2O+': min(_O, _H/2),
    }
    res = rs.as_per_substance_dict(rs.upper_conc_bounds(c0))
    assert res == ref


@requires(parsing_library)
def test_ReactionSystem__identify_equilibria():
    rsys = ReactionSystem.from_string("""
    2 H2 +  O2 -> 2 H2O     ; 1e-3
           H2O -> H+ + OH-  ; 1e-4/55.35
      H+ + OH- -> H2O       ; 1e10
         2 H2O -> 2 H2 + O2
    """)
    assert rsys.identify_equilibria() == [(0, 3), (1, 2)]


@requires(parsing_library)
def test_ReactionSystem__categorize_substances():
    rsys1 = ReactionSystem.from_string("""
    2 H2 +  O2 -> 2 H2O     ; 1e-3
           H2O -> H+ + OH-  ; 1e-4/55.35
      H+ + OH- -> H2O       ; 1e10
         2 H2O -> 2 H2 + O2
    """)
    assert all(not s for s in rsys1.categorize_substances().values())

    rsys2 = ReactionSystem.from_string('\n'.join(['2 NH3 -> N2 + 3 H2', 'N2H4 -> N2 +   2  H2']))
    assert rsys2.categorize_substances() == dict(accumulated={'N2', 'H2'}, depleted={'NH3', 'N2H4'},
                                                 unaffected=set(), nonparticipating=set())

    rsys3 = ReactionSystem.from_string("H+ + OH- -> H2O; 'kf'")
    assert rsys3.categorize_substances() == dict(accumulated={'H2O'}, depleted={'H+', 'OH-'},
                                                 unaffected=set(), nonparticipating=set())

    rsys4 = ReactionSystem([Reaction({'H2': 2, 'O2': 1}, {'H2O': 2})], 'H2 O2 H2O N2 Ar')
    assert rsys4.categorize_substances() == dict(accumulated={'H2O'}, depleted={'H2', 'O2'},
                                                 unaffected=set(), nonparticipating={'N2', 'Ar'})

    rsys5 = ReactionSystem.from_string("""
    A -> B; MassAction(unique_keys=('k1',))
    B + C -> A + C; MassAction(unique_keys=('k2',))
    2 B -> B + C; MassAction(unique_keys=('k3',))
    """, substance_factory=lambda formula: Substance(formula))
    assert rsys5.categorize_substances() == dict(accumulated={'C'}, depleted=set(),
                                                 unaffected=set(), nonparticipating=set())

    rsys6 = ReactionSystem.from_string("""H2O2 + Fe+3 + (H2O2) -> 2 H2O + O2 + Fe+3""")
    assert rsys6.rxns[0].order() == 2  # the additional H2O2 within parenthesis
    assert rsys6.categorize_substances() == dict(accumulated={'H2O', 'O2'}, depleted={'H2O2'},
                                                 unaffected={'Fe+3'}, nonparticipating=set())


@requires(parsing_library)
def test_ReactionSystem__split():
    a = """
    2 H2 +  O2 -> 2 H2O     ; 1e-3
           H2O -> H+ + OH-  ; 1e-4/55.35
      H+ + OH- -> H2O       ; 1e10
        2 H2O  -> 2 H2 + O2"""
    b = """
        2 N    -> N2"""
    c = """
        2 ClBr -> Cl2 + Br2
    """
    rsys1 = ReactionSystem.from_string(a+b+c)
    res = rsys1.split()
    ref = list(map(ReactionSystem.from_string, [a, b, c]))
    for rs in chain(res, ref):
        rs.sort_substances_inplace()
    res1a, res1b, res1c = res
    ref1a, ref1b, ref1c = ref
    assert res1a == ref1a
    assert res1b == ref1b
    assert res1c == ref1c
    assert res1c != ref1a
    assert rsys1.categorize_substances() == dict(
        accumulated={'N2', 'Cl2', 'Br2'}, depleted={'N', 'ClBr'},
        unaffected=set(), nonparticipating=set())


def test_ReactionSystem__subset():
    r1 = Reaction({'NH3': 2}, {'N2': 1, 'H2': 3})
    r2 = Reaction({'N2H4': 1}, {'N2': 1, 'H2': 2})
    rs1 = ReactionSystem([r1, r2])
    rs2 = rs1.subset(lambda r: 'N2H4' in r.keys())
    assert len(rs1.rxns) == 2 and len(rs2.rxns) == 1
    assert rs2 == ReactionSystem([r2])



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)
import re


try:
    import numpy as np
except ImportError:
    np = None


# See https://github.com/python-quantities/python-quantities/pull/112
def format_units_html(udict, font='%s', mult=r'&sdot;', paren=False):
    '''
    Replace the units string provided with an equivalent html string.

    Exponentiation (m**2) will be replaced with superscripts (m<sup>2</sup>})

    No formating is done, change `font` argument to e.g.:
    '<span style="color: #0000a0">%s</span>' to have text be colored blue.

    Multiplication (*) are replaced with the symbol specified by the mult
    argument. By default this is the latex &sdot; symbol.  Other useful options
    may be '' or '*'.

    If paren=True, encapsulate the string in '(' and ')'

    '''
    from quantities.markup import format_units
    res = format_units(udict)
    if res.startswith('(') and res.endswith(')'):
        # Compound Unit
        compound = True
    else:
        # Not a compound unit
        compound = False
    # Replace exponentiation (**exp) with ^{exp}
    res = re.sub(r'\*{2,2}(?P<exp>\d+)', r'<sup>\g<exp></sup>', res)
    # Remove multiplication signs
    res = re.sub(r'\*', mult, res)
    if paren and not compound:
        res = '(%s)' % res
    res = font % res
    return res


def _patch_pow0_py35(pq):
    try:
        pq.metre**0
    except Exception:
        pq.quantity.Quantity.__pow__ = pq.quantity.check_uniform(lambda self, other: np.power(self, other))


def _patch_quantities(pq):
    # See https://github.com/python-quantities/python-quantities/pull/112
    if not hasattr(pq.dimensionality.Dimensionality, 'html'):
        pq.dimensionality.Dimensionality.html = property(lambda self: format_units_html(self))

    # See https://github.com/python-quantities/python-quantities/pull/116
    a = pq.UncertainQuantity([1, 2], pq.m, [.1, .2])
    if (-a).uncertainty[0] != a.uncertainty[0]:
        pq.UncertainQuantity.__neg__ = lambda self: self*-1
    a = pq.UncertainQuantity([1, 2], pq.m, [.1, .2])
    assert (-a).uncertainty[0] == (a*-1).uncertainty[0]

    # See https://github.com/python-quantities/python-quantities/pull/126
    _patch_pow0_py35(pq)
    assert (3*pq.m)**0 == 1*pq.dimensionless



# -*- coding: utf-8 -*-

from __future__ import (absolute_import, division, print_function)

import math

from ..eyring import (
    eyring_equation, EyringParam, EyringParamWithUnits
)
from chempy.util.testing import requires
from chempy.units import allclose, units_library, default_units as u


_kB_over_h = 2.083664399411865234375e10
_R = 8.314472


def test_eyring_equation():
    dH, dS = 40e3, 1e2
    T = 123.45
    ref = _kB_over_h * T * math.exp(dS/_R) * math.exp(-dH/_R/T)
    res = eyring_equation(dH, dS, T)
    assert abs((res - ref)/ref) < 1e-10


def test_EyringParam():
    T = 273.15
    k = EyringParam(40e3, 1e2)(T)
    ref = _kB_over_h * T * math.exp(1e2/_R) * math.exp(-40e3/_R/T)
    assert abs((k - ref)/ref) < 1e-7


@requires(units_library)
def test_EyringParamWithUnits():
    mol = u.mol
    J = u.joule
    K = u.kelvin
    T = 273.15*K
    k = EyringParamWithUnits(42e3 * J/mol, 123*J/mol/K)(T)
    ref = _kB_over_h * 273.15 * math.exp(123/_R) * math.exp(-42e3/_R/273.15)
    ref /= u.second
    assert allclose(k, ref)



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

from itertools import chain
from .periodic import groups, symbols, names

_anions = {
    'F-': 'fluoride',
    'Cl-': 'chloride',
    'Br-': 'bromide',
    'I-': 'iodide',
    'OH-': 'hydroxide',
    'CN-': 'cyanide',
    'SCN-': 'thiocyanate',
    'CO3-2': 'carbonate',
    'C2O4-2': 'oxalate',
    'HCO3-': 'hydrogencarbonate',
    'NO3-': 'nitrate',
    'NO2-': 'nitrite',
    'PO4-3': 'phospahte',
    'HPO4-2': 'hydrogenphospahte',
    'H2PO4-': 'dihydrogenphospahte',
    'P-3': 'phosphide',
    'SO4-2': 'sulphate',
    'HSO4-': 'hydrogensulphate',
    'SO3-2': 'sulphite',
    'HSO3-': 'hydrogensulphite',
    'S-2': 'sulfide',
    'ClO-': 'hypochlorite',
    'ClO2-': 'chlorite',
    'ClO3-': 'chlorate',
    'ClO4-': 'perchlorate',
    'CrO4-2': 'chromate(VI)',
    'Cr2O7-2': 'dichromate(VI)',
    'MnO4-2': 'manganate(VI)',
    'MnO4-': 'permanganate(VII)',
    'FeO4-2': 'ferrate(VI)',
    'OsO4-2': 'osmate(VI)',
    'Bo3-3': 'borate',
    'BiO3-': 'bismuthate(V)',
}
_cations = {
    'H3O+': 'hydronium',
}
_cation_oxidation_states = {  # This needs to be reviewed, just from the top of my head
    'Cr': (2, 3),
    'Fe': (2, 3),
    'Mn': (2,),
    'Co': (2, 3),
    'Ni': (2, 3),
    'Cu': (1, 2, 3),
    'Ag': (1, 2),
    'Au': (3,),
    'Zn': (2,),
    'Cd': (2,),
    'Hg': (1, 2),  # Tricky: Hg2+2
    'Al': (3,),
    'Ga': (3,),
    'In': (3,),
    'Tl': (1, 3),
    'Sn': (2, 4),
    'Pb': (2, 4),
    'Bi': (3,),
    'Sb': (3,),
}

_alkali = [
    (symbols[n]+'+', names[n].lower()) for n in groups[1]
]
_alkaline_earth = [
    (symbols[n]+'+2', names[n].lower()) for n in groups[2]
]
_all_names = dict(chain(_alkali, _alkaline_earth, _anions.items()))


def name(ion):
    return _all_names[ion]


def ions_from_formula(formula):
    """
    This will be working examples eventually:

    #>>> ions_from_formula('NaCl') == {'Na+': 1, 'Cl-': 1}
    #True
    #>>> ions_from_formula('Fe(NO3)3') == {'Fe+3': 1, 'NO3-': 3}
    #True
    #>>> ions_from_formula('FeSO4') == {'Fe+2': 1, 'SO4-2': 1}
    #True
    #>>> ions_from_formula('(NH4)3PO4') == {'NH4+': 3, 'PO4-3': 1}
    #True
    #>>> ions_from_formula('KAl(SO4)2.11H2O') == {'K+': 1, 'Al+3': 1, 'SO4-2': 2}
    #True

    """
    pass



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

import math

from chempy.chemistry import Equilibrium
from chempy.util._expr import Expr
from chempy.util.testing import requires
from chempy.units import allclose, units_library, Backend, default_units as du, default_constants as dc


from ..expressions import MassActionEq, GibbsEqConst


@requires('sympy')
def test_MassActionEq_symbolic():
    import sympy as sp
    K, A, B, C = sp.symbols('K A B C')
    mae = MassActionEq([K])
    eq = Equilibrium({'A'}, {'B', 'C'})
    expr = mae.equilibrium_equation({'A': A, 'B': B, 'C': C}, equilibrium=eq)
    assert expr - K + B*C/A == 0


def test_GibbsEqConst():
    R, T = 8.314, 298.15
    dH, dS = -4e3, 16
    gee = GibbsEqConst([dH/R, dS/R])
    ref = math.exp(-(dH - T*dS)/(R*T))
    assert abs((gee({'temperature': T}) - ref)/ref) < 1e-14


def _gibbs(args, T, R, backend):
    H, S, Cp, Tref = args
    H2 = H + Cp*(T - Tref)
    S2 = S + Cp*backend.log(T/Tref)
    return backend.exp(-(H2 - T*S2)/(R*T))


def test_custom_gibbs():
    R, T = 8.314, 298.15
    dH, dS = -4e3, 16
    MyGibbs = MassActionEq.from_callback(_gibbs, parameter_keys=('temperature', 'R'),
                                         argument_names=('H', 'S', 'Cp', 'Tref'))
    dCp = 123.45
    Tref = 242
    gee2 = MyGibbs([dH, dS, dCp, Tref])
    dH2 = dH + dCp*(T - Tref)
    dS2 = dS + dCp*math.log(T/Tref)
    ref2 = math.exp(-(dH2 - T*dS2)/(R*T))
    assert abs((gee2.eq_const({'temperature': T, 'R': R}) - ref2)/ref2) < 1e-14


def test_GibbsEqConst__unique_keys():
    R, T = 8.314, 298.15
    dH, dS = -4e3, 16
    gee = GibbsEqConst(unique_keys=('dH1', 'dS1'))
    ref = math.exp(-(dH - T*dS)/(R*T))
    assert abs((gee.eq_const({'temperature': T, 'dH1': dH/R, 'dS1': dS/R}) - ref)/ref) < 1e-14


@requires('sympy')
def test_GibbsEqConst__latex():
    import sympy
    DH, DS, R, T = sympy.symbols('\Delta\ H \Delta\ S R T')
    gee = GibbsEqConst([DH/R, DS/R])
    res = gee.eq_const({'temperature': T}, backend=sympy)
    ref = sympy.exp(-(DH - T*DS)/(R*T))
    assert (res - ref).simplify() == 0


@requires(units_library)
def test_GibbsEqConst__units():
    R, T = dc.molar_gas_constant, 298.15*du.K
    DH = -4e3 * du.J/du.mol
    DS = 16 * du.J/du.K/du.mol
    be = Backend()
    gee = GibbsEqConst([DH/R, DS/R])
    res = gee.eq_const({'temperature': T}, backend=be)
    ref = be.exp(-(DH - T*DS)/(R*T))
    assert allclose(res, ref)


@requires(units_library)
def test_GibbsEqConst__nested():

    class TExpr(Expr):
        argument_names = ('heat_capacity',)
        parameter_keys = ('energy',)

        def __call__(self, variables, backend=None):
            heat_capacity, = self.all_args(variables, backend=backend)
            energy, = self.all_params(variables, backend=backend)
            return energy/heat_capacity

    R = 8.314 * du.J/du.K/du.mol
    T = TExpr([10.0 * du.J/du.K])
    dH, dS = -4e3 * du.J/du.mol, 16 * du.J/du.K/du.mol
    gee = GibbsEqConst([dH/R, dS/R])
    be = Backend()
    Tref = 298.15 * du.K
    ref = be.exp(-(dH - Tref*dS)/(R*Tref))
    assert be.abs((gee.eq_const({'energy': 2981.5 * du.J, 'temperature': T}, backend=be) - ref)/ref) < 1e-14



# -*- coding: utf-8 -*-

from __future__ import (absolute_import, division, print_function)

import math

from ..arrhenius import (
    arrhenius_equation, ArrheniusParam, ArrheniusParamWithUnits
)
from chempy.util.testing import requires
from chempy.units import default_units, units_library


def test_arrhenius_equation():
    assert abs(arrhenius_equation(3, 831.4472, 100) - 3/2.7182818) < 1e-7


def test_ArrheniusParam():
    k = ArrheniusParam(1e10, 42e3)(273.15)
    ref = 1e10 * math.exp(-42e3/(8.3145*273.15))
    assert abs((k - ref)/ref) < 1e-4


@requires(units_library)
def test_ArrheniusParamWithUnits():
    s = default_units.second
    mol = default_units.mol
    J = default_units.joule
    K = default_units.kelvin
    k = ArrheniusParamWithUnits(1e10/s, 42e3 * J/mol)(273.15*K)
    ref = 1e10/s * math.exp(-42e3/(8.3145*273.15))
    assert abs((k - ref)/ref) < 1e-4



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)
import math


def nernst_potential(ion_conc_out, ion_conc_in, charge, T,
                     constants=None, units=None, backend=math):
    """
    Calculates the Nernst potential using the Nernst equation for a particular
    ion.

    Parameters
    ----------
    ion_conc_out : float with unit
        Extracellular concentration of ion.
    ion_conc_in : float with unit
        Intracellular concentration of ion.
    charge : integer
        Charge of the ion.
    T : float with unit
        Absolute temperature.
    constants : object (optional, default: None)
        Constant attributes accessed:
            F - Faraday constant
            R - Ideal Gas constant
    units : object (optional, default: None)
        Unit attributes: coulomb, joule, kelvin, mol.
    backend : module (optional, default: math)
        Module used to calculate log using `log` method, can be substituted
        with sympy to get symbolic answers.

    Returns
    -------
    Membrane potential.

    """
    if constants is None:
        F = 96485.33289
        R = 8.3144598
        if units is not None:
            F *= units.coulomb / units.mol
            R *= units.joule / units.kelvin / units.mol
    else:
        F = constants.Faraday_constant
        R = constants.molar_gas_constant

    return (R * T) / (charge * F) * backend.log(ion_conc_out / ion_conc_in)



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

from collections import defaultdict

try:
    import numpy as np
except ImportError:
    np = None
import pytest

from ..util.testing import requires
from ..units import (
    allclose, concatenate, get_derived_unit, is_unitless, linspace, logspace_from_lin,
    SI_base_registry, unitless_in_registry, format_string, get_physical_dimensionality,
    to_unitless, magnitude, default_unit_in_registry, Backend, latex_of_unit,
    unit_of, unit_registry_to_human_readable, units_library, simplified, uniform,
    unit_registry_from_human_readable, _sum, UncertainQuantity, compare_equality,
    default_units as u, patched_numpy as pnp, default_constants as dc
)


@requires(units_library)
def test_default_units():
    u.metre
    u.second
    u.hour
    u.decimetre
    u.mole
    u.kilogram
    u.ampere
    u.kelvin
    u.candela
    u.molar
    u.per100eV
    u.joule
    u.gray
    u.eV
    u.MeV
    u.metre
    u.decimetre
    u.centimetre
    u.micrometre
    u.nanometre
    u.gram
    u.molar
    u.hour
    u.perMolar_perSecond
    u.per100eV
    u.umol
    u.umol_per_J


@requires(units_library)
def test_allclose():
    assert allclose(42, 42)
    assert allclose(42*u.meter, 0.042*u.km)
    assert not allclose(42, 43)
    assert not allclose(42, 42*u.meter)
    assert not allclose(42, 43*u.meter)
    assert not allclose(42*u.meter, 42)

    a = np.linspace(2, 3)*u.second
    b = np.linspace(2/3600., 3/3600.)*u.hour
    assert allclose(a, b)
    assert allclose([3600*u.second, 2*u.metre/u.hour],
                    [1*u.hour, 2/3600*u.metre/u.second])
    c1 = [[3000, 4000], [3000, 4000]]*u.mol/u.metre**3
    c2 = [[3000, 4000], [436.2, 5281.89]]*u.mol/u.metre**3
    assert not allclose(c1, c2)
    assert allclose(0*u.second, 0*u.second)

    # Possibly allow comparison with scalars in future (broadcasting):
    # assert allclose(2, [2, 2])
    # assert allclose([2, 2], 2)

    # assert not allclose(2, [2, 3])
    # assert not allclose([2, 3], 2)

    # assert allclose(2*u.second, [2, 2]*u.second)
    # assert allclose([2, 2]*u.second, 2*u.second)

    # assert not allclose(2*u.second, [2, 3]*u.second)
    # assert not allclose([2, 3]*u.second, 2*u.second)


@requires(units_library)
def test_is_unitless():
    assert not is_unitless(1*u.second)
    assert is_unitless(1)
    assert is_unitless({'a': 1, 'b': 2.0})
    assert not is_unitless({'a': 2, 'b': 5.0*u.second, 'c': 3})
    assert is_unitless(7*u.molar/u.mole*u.dm3)
    assert is_unitless([2, 3, 4])
    assert not is_unitless([2*u.m, 3*u.m])
    assert not is_unitless([3, 4*u.m])


@requires(units_library)
def test_unit_of():
    assert compare_equality(unit_of(0.1*u.metre/u.second), u.metre/u.second)
    assert not compare_equality(unit_of(0.1*u.metre/u.second), u.kilometre/u.second)
    assert compare_equality(unit_of(7), 1)
    assert unit_of(u.gray).dimensionality == u.gray.dimensionality
    ref = (u.joule/u.kg).simplified.dimensionality
    assert unit_of(u.gray, simplified=True).dimensionality == ref

    assert compare_equality(unit_of(dict(foo=3*u.molar, bar=2*u.molar)), u.molar)
    assert not compare_equality(unit_of(dict(foo=3*u.molar, bar=2*u.molar)), u.second)
    with pytest.raises(Exception):
        unit_of(dict(foo=3*u.molar, bar=2*u.second))
    assert not compare_equality(unit_of(dict(foo=3*u.molar, bar=2*u.molar)), u.mol/u.metre**3)


@requires(units_library)
def test_to_unitless():
    dm = u.decimetre
    vals = [1.0*dm, 2.0*dm]
    result = to_unitless(vals, u.metre)
    assert result[0] == 0.1
    assert result[1] == 0.2
    with pytest.raises(Exception):
        to_unitless([42, 43], u.metre)

    vals = [1.0, 2.0]*dm
    result = to_unitless(vals, u.metre)
    assert result[0] == 0.1
    assert result[1] == 0.2

    length_unit = 1000*u.metre
    result = to_unitless(1.0*u.metre, length_unit)
    assert abs(result - 1e-3) < 1e-12

    amount_unit = 1e-9  # nano
    assert abs(to_unitless(1.0, amount_unit) - 1e9) < 1e-6
    assert abs(to_unitless(3/(u.second*u.molar),
                           u.metre**3/u.mole/u.second) - 3e-3) < 1e-12
    assert abs(to_unitless(2*u.dm3, u.cm3) - 2000) < 1e-12
    assert abs(to_unitless(2*u.m3, u.dm3) - 2000) < 1e-12
    assert (float(to_unitless(UncertainQuantity(2, u.dm3, .3), u.cm3)) - 2000) < 1e-12

    g1 = UncertainQuantity(4.46, u.per100eV, 0)
    g_unit = get_derived_unit(SI_base_registry, 'radiolytic_yield')
    assert abs(to_unitless(g1, g_unit) - 4.46 * 1.036e-7) < 1e-9
    g2 = UncertainQuantity(-4.46, u.per100eV, 0)
    assert abs(to_unitless(-g2, g_unit) - 4.46 * 1.036e-7) < 1e-9

    vals = np.array([1.*dm, 2.*dm], dtype=object)
    result = to_unitless(vals, u.metre)
    assert result[0] == 0.1
    assert result[1] == 0.2

    one_billionth_molar_in_nanomolar = to_unitless(1e-9*u.molar, u.nanomolar)
    assert one_billionth_molar_in_nanomolar == 1


@requires(units_library)
def test_UncertainQuantity():
    a = UncertainQuantity([1, 2], u.m, [.1, .2])
    assert a[1] == [2.]*u.m
    assert (-a)[0] == [-1.]*u.m
    assert (-a).uncertainty[0] == [0.1]*u.m
    assert (-a)[0] == (a*-1)[0]
    assert (-a).uncertainty[0] == (a*-1).uncertainty[0]


@requires(units_library, 'sympy')
def test_to_unitless__sympy():
    import sympy as sp
    assert sp.cos(to_unitless(sp.pi)) == -1
    with pytest.raises(AttributeError):
        to_unitless(sp.pi, u.second)


@requires(units_library)
def test_linspace():
    ls = linspace(2*u.second, 3*u.second)
    assert abs(to_unitless(ls[0], u.hour) - 2/3600.) < 1e-15


@requires(units_library)
def test_logspace_from_lin():
    ls = logspace_from_lin(2*u.second, 3*u.second)
    assert abs(to_unitless(ls[0], u.hour) - 2/3600.) < 1e-15
    assert abs(to_unitless(ls[-1], u.hour) - 3/3600.) < 1e-15


@requires(units_library)
def test_get_derived_unit():
    registry = SI_base_registry.copy()
    registry['length'] = 1e-1*registry['length']
    conc_unit = get_derived_unit(registry, 'concentration')
    dm = u.decimetre
    assert abs(conc_unit - 1*u.mole/(dm**3)) < 1e-12*u.mole/(dm**3)

    registry = defaultdict(lambda: 1)
    registry['amount'] = 1e-9  # nano
    assert abs(to_unitless(1.0, get_derived_unit(
        registry, 'concentration')) - 1e9) < 1e-6


@requires(units_library)
def test_unit_registry_to_human_readable():
    # Not as much human readable as JSON serializable...
    d = defaultdict(lambda: 1)
    assert unit_registry_to_human_readable(d) == dict(
        (x, (1, 1)) for x in SI_base_registry.keys())

    ur = {
        'length': 1e3*u.metre,
        'mass': 1e-2*u.kilogram,
        'time': 1e4*u.second,
        'current': 1e-1*u.ampere,
        'temperature': 1e1*u.kelvin,
        'luminous_intensity': 1e-3*u.candela,
        'amount': 1e4*u.mole
    }
    assert unit_registry_to_human_readable(ur) == {
        'length': (1e3, 'm'),
        'mass': (1e-2, 'kg'),
        'time': (1e4, 's'),
        'current': (1e-1, 'A'),
        'temperature': (1e1, 'K'),
        'luminous_intensity': (1e-3, 'cd'),
        'amount': (1e4, 'mol')
    }
    assert unit_registry_to_human_readable(ur) != {
        'length': (1e2, 'm'),
        'mass': (1e-2, 'kg'),
        'time': (1e4, 's'),
        'current': (1e-1, 'A'),
        'temperature': (1e1, 'K'),
        'luminous_intensity': (1e-3, 'cd'),
        'amount': (1e4, 'mol')
    }


@requires(units_library)
def test_unit_registry_from_human_readable():
    hr = unit_registry_to_human_readable(defaultdict(lambda: 1))
    assert hr == dict((x, (1, 1)) for x in SI_base_registry.keys())
    ur = unit_registry_from_human_readable(hr)
    assert ur == dict((x, 1) for x in SI_base_registry.keys())

    hr = unit_registry_to_human_readable(SI_base_registry)
    assert hr == {
        'length': (1.0, 'm'),
        'mass': (1.0, 'kg'),
        'time': (1.0, 's'),
        'current': (1.0, 'A'),
        'temperature': (1.0, 'K'),
        'luminous_intensity': (1.0, 'cd'),
        'amount': (1.0, 'mol')
    }
    ur = unit_registry_from_human_readable(hr)
    assert ur == SI_base_registry

    ur = unit_registry_from_human_readable({
        'length': (1.0, 'm'),
        'mass': (1.0, 'kg'),
        'time': (1.0, 's'),
        'current': (1.0, 'A'),
        'temperature': (1.0, 'K'),
        'luminous_intensity': (1.0, 'cd'),
        'amount': (1.0, 'mol')
    })
    assert ur == {
        'length': u.metre,
        'mass': u.kilogram,
        'time': u.second,
        'current': u.ampere,
        'temperature': u.kelvin,
        'luminous_intensity': u.candela,
        'amount': u.mole
    }

    ur = unit_registry_from_human_readable({
        'length': (1e3, 'm'),
        'mass': (1e-2, 'kg'),
        'time': (1e4, 's'),
        'current': (1e-1, 'A'),
        'temperature': (1e1, 'K'),
        'luminous_intensity': (1e-3, 'cd'),
        'amount': (1e4, 'mol')
    })
    assert ur == {
        'length': 1e3*u.metre,
        'mass': 1e-2*u.kilogram,
        'time': 1e4*u.second,
        'current': 1e-1*u.ampere,
        'temperature': 1e1*u.kelvin,
        'luminous_intensity': 1e-3*u.candela,
        'amount': 1e4*u.mole
    }

    assert ur != {
        'length': 1e2*u.metre,
        'mass': 1e-3*u.kilogram,
        'time': 1e2*u.second,
        'current': 1e-2*u.ampere,
        'temperature': 1e0*u.kelvin,
        'luminous_intensity': 1e-2*u.candela,
        'amount': 1e3*u.mole
    }


@requires(units_library)
def test_unitless_in_registry():
    mag = magnitude(unitless_in_registry(3*u.per100eV, SI_base_registry))
    ref = 3*1.0364268834527753e-07
    assert abs(mag - ref) < 1e-14
    ul = unitless_in_registry([3*u.per100eV, 5*u.mol/u.J], SI_base_registry)
    assert allclose(ul, [ref, 5], rtol=1e-6)


@requires(units_library)
def test_compare_equality():
    assert compare_equality(3*u.m, 3*u.m)
    assert compare_equality(3*u.m, 3e-3*u.km)
    assert compare_equality(3e+3*u.mm, 3*u.m)
    assert not compare_equality(3*u.m, 2*u.m)
    assert not compare_equality(3*u.m, 3*u.s)
    assert not compare_equality(3*u.m, 3*u.m**2)
    assert not compare_equality(3*u.m, np.array(3))
    assert not compare_equality(np.array(3), 3*u.m)
    assert compare_equality([3, None], [3, None])
    assert not compare_equality([3, None, 3], [3, None, None])
    assert not compare_equality([None, None, 3], [None, None, 2])
    assert compare_equality([3*u.m, None], [3, None])
    assert not compare_equality([3*u.m, None], [3*u.km, None])


@requires(units_library)
def test_get_physical_dimensionality():
    assert get_physical_dimensionality(3*u.mole) == {'amount': 1}
    assert get_physical_dimensionality([3*u.mole]) == {'amount': 1}
    assert get_physical_dimensionality(42) == {}


@requires(units_library)
def test_default_unit_in_registry():
    mol_per_m3 = default_unit_in_registry(3*u.molar, SI_base_registry)
    assert magnitude(mol_per_m3) == 1
    assert mol_per_m3 == u.mole/u.metre**3

    assert default_unit_in_registry(3, SI_base_registry) == 1
    assert default_unit_in_registry(3.0, SI_base_registry) == 1


@requires(units_library)
def test__sum():
    # sum() does not work here...
    assert (_sum([0.1*u.metre, 1*u.decimetre]) - 2*u.decimetre)/u.metre == 0


@requires(units_library)
def test_Backend():
    b = Backend()
    with pytest.raises(ValueError):
        b.exp(-3*u.metre)
    assert abs(b.exp(1234*u.metre/u.kilometre) - b.exp(1.234)) < 1e-14


@requires(units_library, 'numpy')
def test_Backend__numpy():
    import numpy as np
    b = Backend(np)
    b.sum([1000*u.metre/u.kilometre, 1], axis=0) == 2.0

    with pytest.raises(AttributeError):
        b.Piecewise


@requires('sympy')
def test_Backend__sympy():
    b = Backend('sympy')
    b.sin(b.pi) == 0

    with pytest.raises(AttributeError):
        b.min


@requires(units_library)
def test_format_string():
    assert format_string(3*u.gram/u.metre**2) == ('3', 'g/m**2')
    assert format_string(3*u.gram/u.metre**2, tex=True) == ('3', r'\mathrm{\frac{g}{m^{2}}}')


@requires(units_library)
def test_joule_html():
    joule_htm = 'kg&sdot;m<sup>2</sup>/s<sup>2</sup>'
    joule = u.J.dimensionality.simplified
    assert joule.html == joule_htm


@requires(units_library)
def test_latex_of_unit():
    assert latex_of_unit(u.gram/u.metre**2) == r'\mathrm{\frac{g}{m^{2}}}'


@requires(units_library)
def test_concatenate():
    a = [1, 2]*u.metre
    b = [2, 3]*u.mm
    ref = [1, 2, 2e-3, 3e-3]*u.metre
    assert allclose(concatenate((a, b)), ref)


@requires(units_library)
def test_pow0():
    a = [1, 2]*u.metre
    b = a**0
    assert allclose(b, [1, 1])

    c = a**2
    assert allclose(c, [1, 4]*u.m**2)


@requires(units_library)
def test_tile():
    a = [2*u.m, 3*u.km]
    assert allclose(pnp.tile(a, 2), [2*u.m, 3000*u.m, 2e-3*u.km, 3*u.km])


@requires(units_library)
def test_simplified():
    assert allclose(simplified(dc.molar_gas_constant), 8.314*u.J/u.mol/u.K, rtol=2e-3)


@requires(units_library)
def test_polyfit_polyval():
    p1 = pnp.polyfit([0, 1, 2], [0, 1, 4], 2)
    assert allclose(p1, [1, 0, 0], atol=1e-14)
    assert allclose(pnp.polyval(p1, 3), 9)
    assert allclose(pnp.polyval(p1, [4, 5]), [16, 25])

    p2 = pnp.polyfit([0, 1, 2]*u.s, [0, 1, 4]*u.m, 2)
    for _p, _r, _a in zip(p2, [1*u.m/u.s**2, 0*u.m/u.s, 0*u.m],
                          [0*u.m/u.s**2, 1e-15*u.m/u.s, 1e-15*u.m]):
        assert allclose(_p, _r, atol=_a)
    assert allclose(pnp.polyval(p2, 3*u.s), 9*u.m)
    assert allclose(pnp.polyval(p2, [4, 5]*u.s), [16, 25]*u.m)


@requires(units_library)
def test_uniform():
    base = [3*u.km, 200*u.m]
    refs = [np.array([3000, 200]), np.array([3, 0.2])]

    def _check(case, ref):
        assert np.any(np.all(magnitude(uniform(case)) == ref, axis=1))
    _check(base, refs)
    _check(tuple(base), refs)
    keys = 'foo bar'.split()
    assert magnitude(uniform(dict(zip(keys, base)))) in [dict(zip(keys, r)) for r in refs]



# -*- coding: utf-8 -*-

import warnings

try:
    from numpy import any as _any
except ImportError:
    def _any(arg):
        if arg is True:
            return True
        if arg is False:
            return False
        return any(arg)

# Parameters from paper (SI-units):

gamma = 2.063
dgamma = 0.051
D0 = 1.635e-8
TS = 215.05
dD0 = 2.242e-11
dTS = 1.2
low_t_bound = 273.15  # 0 deg C, (m.p. at ambient pressure)
high_t_bound = 373.15  # 100 deg C, (b.p. at ambient pressure)


def water_self_diffusion_coefficient(T=None, units=None, warn=True,
                                     err_mult=None):
    """
    Temperature-dependent self-diffusion coefficient of water.

    Parameters
    ----------
    T : float
        Temperature (default: in Kelvin)
    units : object (optional)
        object with attributes: Kelvin, meter, kilogram
    warn : bool (default: True)
        Emit UserWarning when outside temperature range.
    err_mult : length 2 array_like (default: None)
        Perturb paramaters D0 and TS with err_mult[0]*dD0 and
        err_mult[1]*dTS respectively, where dD0 and dTS are the
        reported uncertainties in the fitted paramters. Useful
        for estimating error in diffusion coefficient.

    References
    ----------
    Temperature-dependent self-diffusion coefficients of water and six selected
        molecular liquids for calibration in accurate 1H NMR PFG measurements
        Manfred Holz, Stefan R. Heila, Antonio Saccob;
        Phys. Chem. Chem. Phys., 2000,2, 4740-4742
        http://pubs.rsc.org/en/Content/ArticleLanding/2000/CP/b005319h
        DOI: 10.1039/B005319H
    """
    if units is None:
        K = 1
        m = 1
        s = 1
    else:
        K = units.Kelvin
        m = units.meter
        s = units.second
    if T is None:
        T = 298.15*K
    _D0 = D0 * m**2 * s**-1
    _TS = TS * K
    if err_mult is not None:
        _dD0 = dD0 * m**2 * s**-1
        _dTS = dTS * K
        _D0 += err_mult[0]*_dD0
        _TS += err_mult[1]*_dTS
    if warn and (_any(T < low_t_bound*K) or _any(T > high_t_bound*K)):
        warnings.warn("Temperature is outside range (0-100 degC)")
    return _D0*((T/_TS) - 1)**gamma

# bibtex format (generated at doi2bib.org):
reference = {
    'doi': '10.1039/b005319h',
    'url': 'http://dx.doi.org/10.1039/B005319H',
    'year ': 2000,
    'publisher': 'Royal Society of Chemistry ({RSC})',
    'volume': 2,
    'number': 20,
    'pages': (4740, 4742),
    'author': 'Manfred Holz and Stefan R. Heil and Antonio Sacco',
    'title': ('Temperature-dependent self-diffusion coefficients of water and'
              ' six selected molecular liquids for calibration in accurate 1H'
              ' {NMR} {PFG} measurements'),
    'journal': 'Phys. Chem. Chem. Phys.'
}



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

from .util.pyutil import AttributeContainer


def get_constant_symbols(Symbol=None):
    if Symbol is None:
        from sympy import Symbol
    consts = [
        ('Faraday_constant', 'F'),
        ('Avogadro_constant', 'N_A'),
        ('vacuum_permittivity', 'epsilon_0'),
        ('Boltzmann_constant', 'k_B'),
        ('pi', 'pi'),
        ('molar_gas_constant', 'R'),
    ]
    return AttributeContainer(**{k: Symbol(v) for k, v in consts})
